/*************************************************************************/
/* Copyright (C) 1995-1997 Marlis Hochbruck, Christian Lubich, Hubert    */
/* Selhofer. All rights reserved.                                        */
/*                                                                       */
/* C code written by Mathias Froehlich.                                  */
/*                                                                       */
/* This code is part of a copyrighted package. For details, see the file */
/* "COPYING" in the top-level directory.                                 */
/*************************************************************************/
/*
 * This code is part of EXP4.
 *
 * This module builds and evaluates the krylov space by sqared coupled two
 * term recursions.
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <math.h>
#include "ftypes.h"
#include "fblas.h"
#include "flapack.h"
#include "exp4.h"
#include "prec.h"
#include "errors.h"

/* #define LOCAL_DEBUG_LEVEL 15 */
#define LOCAL_DEBUG_OFFSET 6
#include "debug.h"

#include "pfuns.h"
#include "work.h"
#include "krylov.h"
#include "scnorm.h"
#include "krystep.h"

#include "cgs.h"

#define MAX_CGS_TRIES  2

#define PLAIN_CGS_4

static finteger inc0 = 0;
static finteger inc1 = 1;

static Complex wzero = ComplexConst(0.0);
static Complex wone  = ComplexConst(1.0);

static int DisableGuessFcn = 0;

void PREC(Exp4SetCGSDisableGuessFcn) (Exp4Boolean Val)
{
  if (Val == Exp4True)
    DisableGuessFcn = 1;
  else
    DisableGuessFcn = 0;
}


extern void flaprintf(FILE *,char *,...);


/* returns the size of workspace required for Exp4AddKrylovDimension */
static void PREC(Exp4SizeCgsWork) (struct Krylov *Kry,size_t *Shared,
      size_t *Resident);

/* distributes dynamic workspace, sets up the base and writes the norm of
 * the start vector */
static WorkSpace PREC(Exp4InitCgs) (WorkSpace Work,Complex *Start,
      Exp4KrylovInitType What,struct Krylov *Kry);

/* does the cgs iteration to get H up to the dimension Guess */
static finteger PREC(Exp4KrylovGetHGuessCgs) (WorkSpace Work,
      Complex *Start,finteger Guess,struct Krylov *Kry);

/* increases the krylov space */
static void PREC(Exp4IncreaseDimensionCgs) (WorkSpace Work,
      Real *ResiualNorm,int *AcceptKonvergence,
      Complex *X,finteger ldX,Complex *Z,finteger ldZ,
      finteger fac_t,Complex *Start,finteger UpTo,struct Krylov *Kry);

/* rebuilds the V base and calculate the linear combination */
static void PREC(Exp4KrylovCombineLinearCgs) (Complex *x,
      finteger ldx,Complex *Start,struct Krylov *Kry,
      Complex *zz,finteger ldz,finteger fac_t,int HaveOldV,
      int ForceRebuild);




/* returns a Cgs Krylov 'Handle' */
void PREC(Exp4FillHandleWithCgs) (struct Krylov *Kry,
      finteger MaxKrylovDim,finteger neq,int HaveOwnH)
{
  Kry->KrylovData[KrylovActive(Kry)].KrylovSizeWorkFun =
    & PREC(Exp4SizeCgsWork);
  Kry->KrylovData[KrylovActive(Kry)].InitKryFun =
    & PREC(Exp4InitCgs);
  if (DisableGuessFcn)
    Kry->KrylovData[KrylovActive(Kry)].KrylovGetHGuessFun = NULL;
  else
    Kry->KrylovData[KrylovActive(Kry)].KrylovGetHGuessFun =
      & PREC(Exp4KrylovGetHGuessCgs);
  Kry->KrylovData[KrylovActive(Kry)].KrylovIncreaseDimensionFun =
    & PREC(Exp4IncreaseDimensionCgs);
  Kry->KrylovData[KrylovActive(Kry)].KrylovCombineLinearFun =
    & PREC(Exp4KrylovCombineLinearCgs);

  KrylovMaxSize(Kry)   = MaxKrylovDim;
  KrylovHBandWith(Kry) = 1;

  CgsVStride(Kry) = 3;

  Exp4Message1((DEBUG_FILE,"Created Cgs Handle: Maximum Dimension = %d\n",
	KrylovMaxSize(Kry)));
}


/* returns the size of workspace required for Exp4AddKrylovDimension */
static void PREC(Exp4SizeCgsWork) (krylov *Kry,size_t *Shared,
      size_t *Resident)
{
  size_t SingleLen = sizeof(Complex)*KrylovNeq(Kry);
  size_t VLen = sizeof(Complex)*KrylovNeq(Kry)*CgsVStride(Kry);
  size_t HWork = sizeof(Complex)*KrylovLdSmall(Kry)*KrylovMaxSize(Kry);

  (*Shared) = 0;
  (*Resident) = 5*GetWorkSpaceSize(SingleLen);
  (*Resident) += GetWorkSpaceSize(VLen);

  if (!KrylovCanRebuild(Kry))
    (*Resident) += GetWorkSpaceSize(HWork);
}

/* defines for building the krylov space */
#define VN(I) (CgsV(Kry,(I)%CgsVStride(Kry)))

#define InvalidateVN(Kry,k)						      \
{									      \
  finteger __LastValidOld = KrylovPrivat(Kry).Cgs.LastValidV;		      \
  finteger __LastValidNew = (k)+1-CgsVStride(Kry);			      \
  KrylovPrivat(Kry).Cgs.LastValidV = max(__LastValidOld,__LastValidNew);      \
}
    
#define RN    (CgsR(Kry))

#define RpN   (CgsRp(Kry))

#define SN    (CgsS(Kry))

#define SpN   (CgsSp(Kry))


static WorkSpace PREC(Exp4InitCgs) (WorkSpace Work,Complex *Start,
      Exp4KrylovInitType What,struct Krylov *Kry)
{
  switch (What) {
  case Build:
    {
      if (!KrylovCanRebuild(Kry))
      {
	size_t HWork = sizeof(Complex)*
	  KrylovLdSmall(Kry)*KrylovMaxSize(Kry);
	DistributeWorkspace(KrylovH(Kry),HWork,Work);
      }

      {
	int NVec = (GetRemainingWorkSpaceSize(Work)-KrylovReserveSize(Kry))/
	  (sizeof(Complex)*KrylovNeq(Kry));
	size_t Len = max(NVec,8)*KrylovNeq(Kry)*sizeof(Complex);

	CgsVStride(Kry) = max(NVec - 5,3);

	DistributeWorkspace(CgsR(Kry),Len,Work);

	CgsRp(Kry) = CgsR(Kry)+KrylovNeq(Kry);
	CgsS(Kry) = CgsRp(Kry)+KrylovNeq(Kry);
	CgsSp(Kry) = CgsS(Kry)+KrylovNeq(Kry);
	CgsC(Kry) = CgsSp(Kry)+KrylovNeq(Kry);
	CgsVData(Kry) = CgsC(Kry)+KrylovNeq(Kry);

    /*FIXME  use C for V cache as long as possible*/
	Exp4Message1((DEBUG_FILE,"Taking %d vectors for V caching\n",
	      (int)CgsVStride(Kry)));
      }

      /* get the norm of the start vector */
/*       KrylovNormv(Kry) = pnrm2 (&KrylovNeq(Kry),Start,&inc1); */
      
      CgsC_Eq_V0(Kry) = 1;

      CgsReCalculateV(Kry) = 1;
      
      CgsSize(Kry) = 0;

      CgsLastValidV(Kry) = 0;

      CgsClearExactBreakdownCount(Kry);

      break;
    }
  case ReBuild:
    {
      if (!KrylovCanRebuild(Kry))
	ErrorReturn("Can not rebuild unknown krylov space!",
	      EXP4_ERR_TRY_TO_REBUILD_UNKNOWN_KRYLOV_SPACE);
      
      CgsReCalculateV(Kry) = 1;

      CgsVStride(Kry) = 3;

      CgsLastValidV(Kry) = KrylovSize(Kry);

      CgsClearExactBreakdownCount(Kry);

      DistributeWorkspace(CgsVData(Kry),
	    sizeof(Complex)*CgsVStride(Kry)*KrylovNeq(Kry),Work);
      
      DistributeWorkspace(CgsR(Kry),
	    sizeof(Complex)*CgsVStride(Kry)*KrylovNeq(Kry),Work);
      
      break;
    }
  }

  return Work;
}

static int PREC(DoOneCgsStep) (struct Krylov *Kry,Complex *Tmp,
      Complex *Start,Real gamma)
{
  finteger *Neq = &KrylovNeq(Kry);
  Complex whelp;
  Real thelp;

  if (CgsSize(Kry) == 0)
  {
    /* normalize the first vector in the Krylov base */
    pcopy (&KrylovNeq(Kry),Start,&inc1,VN(0),&inc1);
    prscl (&KrylovNeq(Kry),&KrylovNormv(Kry),VN(0),&inc1);
      
    pcopy (&KrylovNeq(Kry),VN(0),&inc1,RN,&inc1);
    pcopy (&KrylovNeq(Kry),VN(0),&inc1,RpN,&inc1);
    pcopy (&KrylovNeq(Kry),VN(0),&inc1,SN,&inc1);
      
    if (CgsC_Eq_V0(Kry))
    {
      Exp4Message2((DEBUG_FILE,"Using V(0) for CgsC(Kry)\n"));
      /* take the first W vector equal to the first V vector */
      pcopy (&KrylovNeq(Kry),VN(0),&inc1,CgsC(Kry),&inc1);
      CgsClearExactBreakdownCount(Kry);
      
      ComplexSetReal(CgsD(Kry),1.0);
    }
    else
    {
      Real norm;
      finteger idist = 2;
      static finteger iseed[4] = {324,567,435,775};
      Exp4Message2((DEBUG_FILE,"Using a random vector for CgsC(Kry)\n"));
      /* take the W start vector to a random vector */
      plarnv (&idist,iseed,&KrylovNeq(Kry),CgsC(Kry));
      /* normalize W */
      norm = pnrm2 (&KrylovNeq(Kry),CgsC(Kry),&inc1);
      prscl (&KrylovNeq(Kry),&norm,CgsC(Kry),&inc1);
      pdot (&CgsD(Kry),&KrylovNeq(Kry),CgsC(Kry),&inc1,VN(0),&inc1);
    }

    /* multiply with the Jacobian store in Sp */
    PREC(Exp4Mv) (*Neq,SpN,RpN,1);
    
    /* Deltap = c'*A*rp */
    pdot (&CgsDp(Kry),Neq,CgsC(Kry),&inc1,SpN,&inc1);
    
    /* Nu = Dp / D */
    pdiv(CgsNu(Kry),CgsDp(Kry),CgsD(Kry));
    
    /* set coresponding entry in H */
    KrylovHent(Kry,0,0) = CgsNu(Kry);
  }
  else
  {
    finteger k = CgsSize(Kry);
    finteger ihelp = k - 1;
    /* initialize the rest of H with zeros */
    pcopy (&ihelp,&wzero,&inc0,&KrylovHent(Kry,0,k),&inc1);
    pcopy (&ihelp,&wzero,&inc0,&KrylovHent(Kry,k,0),&KrylovLdSmall(Kry));
    
    /* Set H entry for gamma */
    ComplexSetReal(KrylovHent(Kry,k,k-1),gamma);
    
    /* add -Nu*S to A*Rp stored in Sp */
    neg(whelp,CgsNu(Kry));
    paxpy(Neq,&whelp,SN,&inc1,SpN,&inc1);
    
    /* scale Sp with 1.0/gamma */
    prscl (Neq,&gamma,SpN,&inc1);
    
    /* scale S with Nu */
    neg(whelp,CgsNu(Kry));
    pscal (Neq,&whelp,SN,&inc1);
    
    /* S = - S*Nu + gamma*Sp */
    praxpy (Neq,&gamma,SpN,&inc1,SN,&inc1);
    
    /* multiply with the Jacobian VN(KrylovSize(Kry)+1) is
     * temporary workspace here */
    PREC(Exp4Mv) (*Neq,Tmp,SN,0);
    
    mul(whelp,CgsNu(Kry),CgsNu(Kry));
    rdiv(whelp,whelp,gamma*gamma);
    pscal (Neq,&whelp,RN,&inc1);
    
    thelp = 1.0/(gamma*gamma);
    praxpy (Neq,&thelp,Tmp,&inc1,RN,&inc1);
    
    /* delta = c'*r */
    pdot (&CgsD(Kry),Neq,CgsC(Kry),&inc1,RN,&inc1);
    
    if (ComplexIsZero(CgsDp(Kry))) return 1;
    
    /* mu = gamma*delta/deltap */
    pdiv(CgsMu(Kry),CgsD(Kry),CgsDp(Kry));
    rmul(CgsMu(Kry),gamma,CgsMu(Kry));
    
    /* H(k,k+1) = mu*nu */
    mul(KrylovHent(Kry,k-1,k),CgsMu(Kry),CgsNu(Kry));
    
    pcopy (Neq,RN,&inc1,SN,&inc1);
    
    neg(whelp,CgsMu(Kry));
    paxpy(Neq,&whelp,SpN,&inc1,SN,&inc1);
    
    mul(whelp,CgsMu(Kry),CgsMu(Kry));
    pscal (Neq,&whelp,RpN,&inc1);
    
    neg(whelp,CgsMu(Kry));
    paxpy(Neq,&whelp,SpN,&inc1,RpN,&inc1);
    
    thelp = 1.0;
    praxpy(Neq,&thelp,SN,&inc1,RpN,&inc1);
    
    /* multiply with the Jacobian */
    PREC(Exp4Mv) (*Neq,SpN,RpN,0);
    
    /* deltap = c'*A*rp */
    pdot (&CgsDp(Kry),Neq,CgsC(Kry),&inc1,SpN,&inc1);
    
    if (ComplexIsZero(CgsD(Kry)))  return 1;
    
    pdiv(CgsNu(Kry),CgsDp(Kry),CgsD(Kry));
    
    rmul(whelp,gamma,CgsMu(Kry));
    add(KrylovHent(Kry,k,k),CgsNu(Kry),whelp);
  }
  CgsSize(Kry) ++;

  Exp4Message5((DEBUG_FILE,"Doing Cgs iteration: dimension = %d\n",
	CgsSize(Kry)));

  return 0;
}

static void PREC(DoOneLanczosStep) (struct Krylov *Kry)
{
  finteger *Neq = &KrylovNeq(Kry);
  Complex whelp;

  if (KrylovSize(Kry) == 0)
  {
    InvalidateVN(Kry,1);
    /* udate the lanczos vectors */
    /* multiply with the Jacobian */
    PREC(Exp4Mv) (*Neq,VN(1),VN(0),1);
    
    neg(whelp,KrylovHent(Kry,0,0));
    paxpy(Neq,&whelp,VN(0),&inc1,VN(1),&inc1);
  }
  else
  {
    finteger k = KrylovSize(Kry);
    finteger kp1 = k + 1;
    finteger km1 = k - 1;
    
    /* Do the three term lanczos recursion */
    /* multiply with the Jacobian */
    prscl (&KrylovNeq(Kry),&Re(KrylovHent(Kry,k,km1)),VN(k),&inc1);
    
    InvalidateVN(Kry,kp1);
    PREC(Exp4Mv) (*Neq,VN(kp1),VN(k),1);
    
    neg(whelp,KrylovHent(Kry,km1,k));
    paxpy(Neq,&whelp,VN(km1),&inc1,VN(kp1),&inc1);
    
    neg(whelp,KrylovHent(Kry,k,k));
    paxpy(Neq,&whelp,VN(k),&inc1,VN(kp1),&inc1);
  }

  KrylovSize(Kry) ++;

  Exp4Message5((DEBUG_FILE,"Doing Lanczos iteration: dimension = %d\n",
	  KrylovSize(Kry)));
}

static finteger PREC(Exp4KrylovGetHGuessCgs) (WorkSpace Work,
      Complex *Start,finteger Guess,struct Krylov *Kry)
{
  if (Guess > (CgsVStride(Kry)-1))
  {
    finteger *Neq = &KrylovNeq(Kry);
    /* indicates if an exact breakdown occures */
    int ExactBreakdown = 0;
    
    Exp4Message4((DEBUG_FILE,"Calculating HGuess: dimension = %d\n", Guess));
    /* force calculating of V while calculating the residual */
    CgsReCalculateV(Kry) = 0;

    /* calculate new base vectors up to the dimension UpTo */
    while (CgsSize(Kry) < Guess)
    {
      /* if this point is reached from an exact breakdown, restart the
       * process with a random start vector for the W base */
      if (ExactBreakdown)
      {
	/* reset the krylov space size */
	CgsSize(Kry) = 0;
	KrylovSize(Kry) = 0;
	/* markthe first W vector different from the first V */
	CgsC_Eq_V0(Kry) = 0;
	/* reset the flag for exact breakdown */
	ExactBreakdown = 0;
	/* count this breakdown */
	CgsIncrementExactBreakdownCount(Kry);
	
	Exp4Message1((DEBUG_FILE,"Cgs Process had an unrecoverable breakdown: "
	      "restarting\n"));
      }
      
      InvalidateVN(Kry,1);
      if (PREC(DoOneCgsStep) (Kry,VN(1),Start,sqrt(pnrm2 (Neq,RN,&inc1))))
      {
	ExactBreakdown = 1;
	continue;
      }
    }
  } else 
    Exp4Message4((DEBUG_FILE,"Not calculating HGuess: caching is cheaper!\n"));
  

  Exp4Message4((DEBUG_FILE,"Exit of GetHGuessCgs Guess = %d, KrySize = %d,"
	" CgsSize = %d\n",Guess,KrylovSize(Kry),CgsSize(Kry)));
  
  return CgsSize(Kry);
}

static void PREC(Exp4IncreaseDimensionCgs) (WorkSpace Work,
      Real *ResiualNorm,int *AcceptKonvergence,
      Complex *X,finteger ldX,Complex *Z,finteger ldZ,
      finteger fac_t,Complex *Start,finteger UpTo,krylov *Kry)
{
#define z(I,J)      Z[(I) + ldZ*(J)]

  finteger *Neq = &KrylovNeq(Kry);
  /* indicates if an exact breakdown occures */
  int ExactBreakdown = 0;

  if ((KrylovSize(Kry) == 0) && !CgsReCalculateV(Kry))
  {
    finteger ihelp = fac_t * ldX;
    pcopy (&ihelp,&wzero,&inc0,X,&inc1);
    Exp4Message4((DEBUG_FILE,"Set Result array to zero!\n"));
  }
  
  /* calculate new base vectors up to the dimension UpTo */
  while (CgsSize(Kry) < UpTo)
  {
    CgsReCalculateV(Kry) = 1;
    /* if this point is reached from an exact breakdown, restart the
     * process with a random start vector for the W base */
    if (ExactBreakdown)
    {
      /* reset the krylov space size */
      CgsSize(Kry) = 0;
      KrylovSize(Kry) = 0;
      /* markthe first W vector different from the first V */
      CgsC_Eq_V0(Kry) = 0;
      /* reset the flag for exact breakdown */
      ExactBreakdown = 0;
      /* count this breakdown */
      CgsIncrementExactBreakdownCount(Kry);

      Exp4Message1((DEBUG_FILE,"Cgs Process had an unrecoverable breakdown: "
	    "restarting\n"));
    }

    /* this VN vector is temporary workspace here */
    InvalidateVN(Kry,1+KrylovSize(Kry));
    if (PREC(DoOneCgsStep) (Kry,VN(KrylovSize(Kry)+1),Start,
	  pnrm2 (Neq,VN(KrylovSize(Kry)),&inc1) ))
    {
      ExactBreakdown = 1;
      continue;
    }

    PREC(DoOneLanczosStep) (Kry);

    /* add the rebuilt vector to the result */
    if (CgsReCalculateV(Kry) == 0)
    {
      Exp4Message4((DEBUG_FILE,"Use V(%d) for linear combination\n",
	    KrylovSize(Kry)-1));
      pger (&KrylovNeq(Kry),&fac_t,&wone,VN(KrylovSize(Kry)-1),&inc1,
	    &z(KrylovSize(Kry)-1,0),&ldZ,X,&ldX);
    }
  }

  /* calculate new base vectors up to the dimension UpTo */
  while (KrylovSize(Kry) < UpTo)
  {
    PREC(DoOneLanczosStep) (Kry);

    /* add the rebuilt vector to the result */
    if (CgsReCalculateV(Kry) == 0)
    {
      Exp4Message4((DEBUG_FILE,"Use V(%d) for linear combination\n",
	    KrylovSize(Kry)-1));
      pger (&KrylovNeq(Kry),&fac_t,&wone,VN(KrylovSize(Kry)-1),&inc1,
	    &z(KrylovSize(Kry)-1,0),&ldZ,X,&ldX);
    }
  } /* while (KrylovSize(Kry) < ...) */


  Exp4Message4((DEBUG_FILE,"Exit of AddKrylovDimensionsCgs KrySize = %d,"
	" CgsSize = %d\n",KrylovSize(Kry),CgsSize(Kry)));
/* flaprintf(stdout,"H =\n%ladgee\n",CgsSize(Kry)+1,CgsSize(Kry)+1, */
/*       KrylovH(Kry),KrylovLdSmall(Kry)); */
/* getchar(); */
  {
    finteger k = KrylovSize(Kry);

    (*ResiualNorm) = PREC(Exp4ScaledNorm) (VN(k));
    (*AcceptKonvergence) = ((CgsSize(Kry) - KrylovSize(Kry) - 1) >
	  (KrylovSize(Kry) - (CgsVStride(Kry)-1)));
  }
#undef z
}

/**********************************************************************/
/*                                                                    */
/*                rebuilding the krylov base                          */
/*                                                                    */
/**********************************************************************/

/* the last V vectors which are here from building the krylov space */
#define STOREDVN(I) (CgsV(Kry,(I)%CgsVStride(Kry)))

/* the rebuilding process will be done in the memory for the Cgs vectors */
#define REBUILDVN(I) (CgsR(Kry)+((I)%3)*KrylovNeq(Kry))

static void PREC(Exp4KrylovCombineLinearCgs) (Complex *X,
      finteger ldX,Complex *Start,krylov *Kry,
      Complex *Z_,finteger ldZ,finteger fac_t,int HaveOldV,
      int ForceRebuild)
     /* This routine rebuilds the V vectors from the stored recursion
      * coeficients.
      * If HaveOldV is set to 1 the vectors stored in V are the last V
      * vectors of the Krylov base, so they can be reused. If HaveOldV
      * is set to 0 the whole base must be rebuilt.
      */
{
#define Z(I)      (Z_+(I))
  /* the linearcombination has to be done here if phikry says that the step
     size has changed (ForceRebuild is true) or if the CGS itteration is not
     konverged before kGuess is reached (than CgsReCalculateV is true) */

  if ((ForceRebuild == 1) || (CgsReCalculateV(Kry) == 1))
  {
    /* initialize the array */
    {
      finteger ihelp = fac_t * ldX;
      pcopy (&ihelp,&wzero,&inc0,X,&inc1);
      
      Exp4Message4((DEBUG_FILE,"Set Result array to zero!\n"));
    }
    
    /* take the start vector for the linear combination */
    if (KrylovSize(Kry) >= 1)
    {
      Complex rNormv;
      ComplexSetReal(rNormv,1.0/KrylovNormv(Kry));
      pger(&KrylovNeq(Kry),&fac_t,&rNormv,Start,&inc1,Z(0),&ldZ,X,&ldX);
      Exp4Message4((DEBUG_FILE,"Use V(0) for linear combination\n"));
    }
    
    if (KrylovSize(Kry) > 1)
    {
      /* helper variable */
      Complex whelp;
      /* the last V needed for linear combination */
      finteger LastVToCalculate =  KrylovSize(Kry);
      /* the actual V index to recompute */
      finteger CurrentVIndex = 1;

      /* FIXME ask for last to calculate > 1 */
      
      /* copy and rescale the first vector */
      pcopy (&KrylovNeq(Kry),Start,&inc1,REBUILDVN(0),&inc1);
      prscl (&KrylovNeq(Kry),&KrylovNormv(Kry),REBUILDVN(0),&inc1);
      
      /* if the V workspace holds the last V vectors calculated while building
       * the Cgs base, reuse them (that means only rebuild the first ones) */
      if (HaveOldV)
	LastVToCalculate = min(CgsLastValidV(Kry),LastVToCalculate);
      
      Exp4Message3((DEBUG_FILE,"Set LastVToCalculate to %d\n",
	    LastVToCalculate));
      
      
      /* rebuild all vectors up to LastVToCalculate, the last vectors are in
       * the V workspace */
      while (CurrentVIndex < LastVToCalculate)
      {
	/* the actual column index of H to process */
	finteger HColumn = CurrentVIndex - 1;
	
	/* multiply with the Jacobian */
	PREC(Exp4Mv) (KrylovNeq(Kry),REBUILDVN(CurrentVIndex),
	      REBUILDVN(CurrentVIndex-1),1);
	
	/* apply the beta coeficients */
	if (1 < CurrentVIndex)
	{
	  neg(whelp,KrylovHent(Kry,CurrentVIndex-2,HColumn));
	  paxpy (&KrylovNeq(Kry),&whelp,REBUILDVN(CurrentVIndex-2),&inc1,
		REBUILDVN(CurrentVIndex),&inc1);
	}
	
	neg(whelp,KrylovHent(Kry,HColumn,HColumn));
	paxpy (&KrylovNeq(Kry),&whelp,REBUILDVN(HColumn),&inc1,
	      REBUILDVN(CurrentVIndex),&inc1);
	
	/* normalize the current V (KrylovHent(Kry,k+1,k) is real anyway) */
	prscl (&KrylovNeq(Kry),&Re(KrylovHent(Kry,CurrentVIndex,HColumn)),
	      REBUILDVN(CurrentVIndex),&inc1);
	
	Exp4Message4((DEBUG_FILE,"Use V(%d) for linear combination\n",
	      CurrentVIndex));
	
	/* add the rebuilt vector to the result */
	pger (&KrylovNeq(Kry),&fac_t,&wone,REBUILDVN(CurrentVIndex),&inc1,
	      Z(CurrentVIndex),&ldZ,X,&ldX);
	
	CurrentVIndex++;
      }
      
      /* use the stored (and partially rebuilt) vectors for the
       * linear kombination */
      while (CurrentVIndex < KrylovSize(Kry))
      {
	Exp4Message4((DEBUG_FILE,"Use V(%d) (cached) for linear combination\n",
	      CurrentVIndex));
	
	/* add the rebuilt vector to the result */
	pger (&KrylovNeq(Kry),&fac_t,&wone,STOREDVN(CurrentVIndex),&inc1,
	      Z(CurrentVIndex),&ldZ,X,&ldX);
	CurrentVIndex++;
      }
    }
  }

  CgsSize(Kry) = KrylovSize(Kry);
    
#undef Z
}
#undef VN
