/*************************************************************************/
/* Copyright (C) 1995-1997 Marlis Hochbruck, Christian Lubich, Hubert    */
/* Selhofer. All rights reserved.                                        */
/*                                                                       */
/* C code written by Mathias Froehlich.                                  */
/*                                                                       */
/* This code is part of a copyrighted package. For details, see the file */
/* "COPYING" in the top-level directory.                                 */
/*************************************************************************/

/*
 * This file contains constants and types which are used in the whole
 * Exp4 package.
 * It also containes the funcions which are important for the user.
 */

#ifndef EXP4_H
#define EXP4_H

/* needed for the FILE type */
#include <stdio.h>

#include "ftypes.h"

/* defines maximum size of the output files */
#define MAX_FILE_SIZE 1024*1024*10

/* defines the default maximum number of dimensions for krylov room */
#define DEFAULT_MAX_KRYLOV_DIM	36

/* defines the default maximum block size in the
 * Look - Ahead Lanczos Algorithem */
#define DEFAULT_MAX_BLOCK_SIZE 3

/* defines the default maximum krylov dimension for use with the
 * combined krylov strategy */
#define DEFAULT_MAX_MIX_SIZE  6

/* some usefull stuff */
#ifdef max
#undef max
#endif
#ifdef min
#undef min
#endif
#define max(a,b) (((a)<(b))?(b):(a))
#define min(a,b) (((a)<(b))?(a):(b))

#ifdef  __cplusplus
extern "C" {
#endif

typedef enum {Exp4False = 0, Exp4True = 1 } Exp4Boolean;

typedef void (*SRhsFunc) (finteger neq,fsingle *res,const fsingle *y);
typedef void (*DRhsFunc) (finteger neq,fdouble *res,const fdouble *y);
typedef void (*CRhsFunc) (finteger neq,fsinglecomplex *res,
      const fsinglecomplex *y);
typedef void (*ZRhsFunc) (finteger neq,fdoublecomplex *res,
      const fdoublecomplex *y);
/* This is the prototype for the right hand side function of the
   problem. You must write the result into the vector res. You get the
   actual y value.*/



typedef void (*SFullJac) (finteger neq,fsingle *a,finteger ld,
      const fsingle *y);
typedef void (*DFullJac) (finteger neq,fdouble *a,finteger ld,
      const fdouble *y);
typedef void (*CFullJac) (finteger neq,fsinglecomplex *a,finteger ld,
      const fsinglecomplex *y);
typedef void (*ZFullJac) (finteger neq,fdoublecomplex *a,finteger ld,
      const fdoublecomplex *y);
/* This is a prototype for the Jacobian function. You must write the
   Jacobian into the array a. The matrix must be stored columnwise
   that means that you must write the first column into a[0] up to
   a[neq-1], the second into a[ld] up to a[ld+neq-1] and so on. The
   leading dimension of A is ld. You get the actual y value. (only for
   exp4ex) */



typedef void (*SNewJac) (finteger neq,const fsingle *y);
typedef void (*DNewJac) (finteger neq,const fdouble *y);
typedef void (*CNewJac) (finteger neq,const fsinglecomplex *y);
typedef void (*ZNewJac) (finteger neq,const fdoublecomplex *y);
/* This is the prototype for a function which generates a new Jacobian
   at the point y in any format you prefer. You have to store that
   Jacobian in your code, the Jacobian itself is not referred by exp4.
   You can do that for example as static global variables in your
   module for the functions of the problem. (only for exp4) */



typedef void (*SEvalJac) (finteger neq,fsingle *res,const fsingle *y,
      const fsingle *direction);
typedef void (*DEvalJac) (finteger neq,fdouble *res,const fdouble *y,
      const fdouble *direction);
typedef void (*CEvalJac) (finteger neq,fsinglecomplex *res,
      const fsinglecomplex *y,const fsinglecomplex *direction);
typedef void (*ZEvalJac) (finteger neq,fdoublecomplex *res,
      const fdoublecomplex *y,const fdoublecomplex *direction);
/* This is the prototype for a function which returns the derivative
   in direction 'direction'. You must write the result into the vector
   res. You get the point y where to evaluate the Jacobian and the
   vector direction, to calculate Jacobian*direction.
   You can calculate that derivative by multiplying vect with the
   Jacobian which is calculated by the function dNewJac, or you can
   calculate this without an explicitly given Jacobian if you can do
   so.
   NOTE: the vector y is NOT avaliable if you specified a
   [SDCZ]NewJac function. If you need the actual y value copy it
   into your own workspace! (only for exp4) */


typedef void (*SEvalJacTransp) (finteger neq,fsingle *res,const fsingle *y,
      const fsingle *v);
typedef void (*DEvalJacTransp) (finteger neq,fdouble *res,const fdouble *y,
      const fdouble *v);
typedef void (*CEvalJacTransp) (finteger neq,fsinglecomplex *res,
      const fsinglecomplex *y,const fsinglecomplex *v);
typedef void (*ZEvalJacTransp) (finteger neq,fdoublecomplex *res,
      const fdoublecomplex *y,const fdoublecomplex *v);
/* This is the prototype for a function which evaluates the vector Matrix
   product v'*Jacobian. You must write the result into the vector res.
   You get the point y where to evaluate the Jacobian and the vector v.
   NOTE: the vector y is NOT avaliable if you specified a
   [SDCZ]NewJac function. If you need the actual y value copy it
   into your own workspace! (only for exp4) */



typedef void (*SSolTrait)(long nr, fdouble told, fdouble t,
			  const fsingle *y,unsigned n,int *irtrn);
typedef void (*DSolTrait)(long nr, fdouble told, fdouble t,
			  const fdouble *y,unsigned n,int *irtrn);
typedef void (*CSolTrait)(long nr, fdouble told, fdouble t,
			  const fsinglecomplex *y,unsigned n,int *irtrn);
typedef void (*ZSolTrait)(long nr, fdouble told, fdouble t,
			  const fdoublecomplex *y,unsigned n,int *irtrn);
/* user-defined routine for output similar to solout in dopri5 (see
   http://www.unige.ch/math/folks/hairer/software.html). nr holds the
   step number, told the time of the last step, t the time of the
   current step, y the new solution at time t, and in irtrn you have
   to return the success of the action, that means if you return any
   number < 0 exp4 will stop and exit. */


extern int SExp4(fsingle *t,fsingle tEnd,fsingle *y0,finteger neq,
      SRhsFunc RHSFunc,SNewJac jac,SEvalJac evjac,Exp4Boolean constantJ);
extern int DExp4(fdouble *t,fdouble tEnd,fdouble *y0,finteger neq,
      DRhsFunc RHSFunc,DNewJac jac,DEvalJac evjac,Exp4Boolean constantJ);
extern int CExp4(fsingle *t,fsingle tEnd,fsinglecomplex *y0,finteger neq,
      CRhsFunc RHSFunc,CNewJac jac,CEvalJac evjac,Exp4Boolean constantJ);
extern int ZExp4(fdouble *t,fdouble tEnd,fdoublecomplex *y0,finteger neq,
      ZRhsFunc RHSFunc,ZNewJac jac,ZEvalJac evjac,Exp4Boolean constantJ);
/* The main driver routine Exp4:
   It uses the Krylov methods.

   t          On entry it should contain the initial time of the problem.
              On exit it holds the time of the last calculated step.
              That is normally tEnd, if no error occurred.

   tEnd       Final time.

   y          On entry y should contain the initial value at the
              initial time t. On exit it contains the solution vector
              at the returned time t (which is usually tEnd, if no
              error occurred).

   neq        The number of equations, i.e. the dimension of the system.

   RHSFunc    A pointer to the right hand side function of the ODE.

   jac        A pointer to the function which generates the Jacobian
              of RHSFunc at the given point y, it has to be stored
              somewhere in your module containing your function and
              Jacobian evaluations.
	      The advantage of writing the Jacobian into your module
	      is that you can use YOUR favorite sparse format.
              If it is set to NULL that function is never referred.

   evjac      A pointer to a function which multiplies a vector with the
              Jacobian.

   ConstantJacobian
              A flag which indicates if the Jacobian is constant.
              ConstantJacobian=Exp4False if it is not
	      constant, else ConstantJacobian=Exp4True. */

extern int SExp4Ex(fsingle *t,fsingle tEnd,fsingle *y0,finteger neq,
      SRhsFunc RHSFunc,SFullJac jac,Exp4Boolean constantJ);
extern int DExp4Ex(fdouble *t,fdouble tEnd,fdouble *y0,finteger neq,
      DRhsFunc RHSFunc,DFullJac jac,Exp4Boolean constantJ);
extern int CExp4Ex(fsingle *t,fsingle tEnd,fsinglecomplex *y0,finteger neq,
      CRhsFunc RHSFunc,CFullJac jac,Exp4Boolean constantJ);
extern int ZExp4Ex(fdouble *t,fdouble tEnd,fdoublecomplex *y0,finteger neq,
      ZRhsFunc RHSFunc,ZFullJac jac,Exp4Boolean constantJ);
/*
 * The main driver routine Exp4:
 * It does not use Krylov methods.
 *
 * t          On entry it should contain the initial time of the problem.
 *            On exit it holds the time of the last calculated step.
 *            That is normally tEnd, if not an error occurred.
 *
 * tEnd       Final time.
 *
 * y          On entry y should contain the initial value at the
 *            initial time t. On exit it contains the solution vector
 *            at the returned time t (which is usually tEnd, if no
 *            error occurred).
 *
 * neq        The number of equations, i.e. the dimension of the system.
 *
 * RHSFunc    A pointer to the right hand side function of the ODE.
 *
 * jac        A pointer to the function which generates the Jacobian of RHSFunc
 *            at the given point y. The result has to be written in the array
 *            res with the leading dimension ld. This array has to be written
 *            in FORTRAN convention, i.e. columnwise.
 *
 * ConstantJacobian
 *            A flag which indicates if the Jacobian is constant.
 *            ConstantJacobian=Exp4False if it is not
 *            constant, else ConstantJacobian=Exp4True.
 */

int SExpEul (fsingle *t,fsingle tEnd,fsingle *y,finteger neq,SRhsFunc RHSFunc,
      SEvalJac evjac);
int DExpEul (fdouble *t,fdouble tEnd,fdouble *y,finteger neq,DRhsFunc RHSFunc,
      DEvalJac evjac);
int CExpEul (fsingle *t,fsingle tEnd,fsinglecomplex *y,finteger neq,
      CRhsFunc RHSFunc,CEvalJac evjac);
int ZExpEul (fdouble *t,fdouble tEnd,fdoublecomplex *y,finteger neq,
      ZRhsFunc RHSFunc,ZEvalJac evjac);
/*
 * ExpEul is a solver for big systems of stiff and linear ODE's.
 * It is NOT the main topic of this package but it was very easy to write
 * the code using the enviroment of Exp4!
 *
 * Parameters:
 *
 * t          On entry it should containe the initial time of the problem.
 *            On exit it holds the time of the last calculated step.
 *            That is normally tEnd, if no error occured.
 *
 * tEnd       The end time you want to calculate.
 *
 * y          On entry y schould containe the initial values at time t.
 *            On exit it containes the solution vector an time t.
 *
 * neq        The number of equations, i.e. the dimension of Image(RHSFunc).
 *
 * RHSFunc    A pointer to the righthand side function of the ODE.
 *
 * evjac      A pointer to a function which multiplies a vector with the
 *            jacobian.
 */



/*
 * defines Krylov space strategy.
 */
typedef enum {
  Arnoldi,          /* Arnoldi for all three stages */
  ArnoldiOneSpace,  /* Arnoldi for all stages, only one base is stored */
  Lanczos,          /* Lanczos for all three stages */
  LanczosMixed,     /* first stage Lanczos, other Arnoldi */
  LanczosSwitched,
  BiCGS,            /* at the moment experimental ;-) */
  BiCGSMixed,       /* at the moment experimental ;-) */
  BiCGSSwitched,
  BiConjugateGradientSqared,  /* at the moment experimental ;-) */
  BiConjugateGradientSqaredMixed,  /* at the moment experimental ;-) */
  BiConjugateGradientSqaredSwitched
} Exp4KrylovStrategy;

extern void SExp4SetKrylovStrategy (Exp4KrylovStrategy Strategy,...);
extern void DExp4SetKrylovStrategy (Exp4KrylovStrategy Strategy,...);
extern void CExp4SetKrylovStrategy (Exp4KrylovStrategy Strategy,...);
extern void ZExp4SetKrylovStrategy (Exp4KrylovStrategy Strategy,...);
/*
 * Routine to set the Krylov space strategy.
 * Arnoldi:         All three steps are calculated with arnoldi.
 *                  (Default)
 * ArnoldiOneSpace: All three steps are calculated with arnoldi.
 *                  Workspace is only allocated for one Krylov base,
 *                  that means in a rejected step the same krylov space
 *                  must be recalculated.
 * Lanczos:         All three steps are calculated with Lanczos.
 *                  In this case the function for the multiplication of
 *                  the transposed Jacobian with a vector is required
 *                  as second argument (Type [SDCZ]EvalJacTransp).
 *
 * The other Types are experimental, so DO NOT USE YET!
 */

extern void SExp4SetMaxKrylovDim (int dim);
extern void DExp4SetMaxKrylovDim (int dim);
extern void CExp4SetMaxKrylovDim (int dim);
extern void ZExp4SetMaxKrylovDim (int dim);
/*
 * Routine sets the maximum allowed krylov dimension (deafult 36).
 */


extern void SExp4SetMaxLanczosBlockSize(int size);
extern void DExp4SetMaxLanczosBlockSize(int size);
extern void CExp4SetMaxLanczosBlockSize(int size);
extern void ZExp4SetMaxLanczosBlockSize(int size);
/*
 * Routine sets the maximum allowed block size in Look - Ahesd Lanczos
 * (deafult 3).
 * NOTE: This Look - Ahead Lanczos implementation does not support large
 *       blocks (say bigger than 6).
 */

extern void SExp4SetMaxArnoldi23Size(int size);
extern void DExp4SetMaxArnoldi23Size(int size);
extern void CExp4SetMaxArnoldi23Size(int size);
extern void ZExp4SetMaxArnoldi23Size(int size);
/*
 * Routine sets the maximum allowed krylov space dimension of the second
 * and third exp4 stage. This is normally set to
 * 2*(2 + maximum block size in lanzcos).
 * So there is no additional workspace is required (storage for Lanczos
 * vectors is taken for Arnoldi vectors).
 * To increase this dimension without increasing Lanczos block size use
 * this funktion. (deafult 6).
 */


extern void SExp4SetHmax(fsingle HmaxIn);
extern void DExp4SetHmax(fdouble HmaxIn);
extern void CExp4SetHmax(fsingle HmaxIn);
extern void ZExp4SetHmax(fdouble HmaxIn);
/* Routine to set an absolute upper bound to the step size h.
   If Exp4SetHmax is never called no upper bounds are used. */

typedef enum { Scalar, Vector, Default } Exp4TolType;
/* enumeration type for use in Exp4Set[AR]tol, see below */

extern int SExp4SetAtol(Exp4TolType Type, ...);
extern int DExp4SetAtol(Exp4TolType Type, ...);
extern int CExp4SetAtol(Exp4TolType Type, ...);
extern int ZExp4SetAtol(Exp4TolType Type, ...);
/* Routine to set the value Atol.
   If Type is set to Scalar, the next variable should be the scalar atol value
   (if you are calculating in single precision, an fsingle value is expected.
   if you are calculating in double precision, an fdouble value is expected.).
   If Type is set to Vector, the next variable should be the pointer to the
   atol vector of length neq (if you are calculating in single precision, a
   pointer to an fsingle array is expected. if you are calculating in double
   precision, a pointer to an fdouble array is expected.). If you set Type to
   Vector, no copying is done and your vector is taken for Atol (Be careful
   Exp4(Ex) refers to __YOUR__ vector Atol). If Exp4SetAtol is never called the
   default values defined in scnorm (currently Atol = 1e-6) are used. */

extern int SExp4SetRtol(Exp4TolType Type, ...);
extern int DExp4SetRtol(Exp4TolType Type, ...);
extern int CExp4SetRtol(Exp4TolType Type, ...);
extern int ZExp4SetRtol(Exp4TolType Type, ...);
/* Routine to set the value Rtol.
   If Type is set to Scalar, the next variable should be the scalar Rtol value.
   (if you are calculating in single precision, an fsingle value is expected.
   if you are calculating in double precision, an fdouble value is expected.)
   If Type is set to Vector, the next variable should be the pointer to the
   Rtol vector of length neq. (if you are calculating in single precision, a
   pointer to an fsingle array is expected. if you are calculating in double
   precision, a pointer to an fdouble array is expected.) If you set Type to
   Vector, no copying is done and your vector is taken for Rtol (Be careful
   Exp4(Ex) refers to __YOUR__ vector Rtol).
   If Exp4SetRtol is never called the default values defined in scnorm
   (currently Rtol = 1e-3) are used. */

typedef enum { NoSpecial, Dense, AllFile, DenseFile, EndFile, Solout }
Exp4OutputType;
/* enumeration Type for use with Exp4SetOutput, see below */

extern void SExp4SetOutput(Exp4OutputType Type, ...);
extern void DExp4SetOutput(Exp4OutputType Type, ...);
extern void CExp4SetOutput(Exp4OutputType Type, ...);
extern void ZExp4SetOutput(Exp4OutputType Type, ...);
/* Routine to set the Type of output which is written by Exp4(Ex).
   TypeIn can be set to the values defined in the enumeration type above.
   The types are:
   NoSpecial:     No output is generated. The solution at time t is
                  stored in the vector y returned by the main
                  routine Exp4(Ex).
   Dense:         Dense output is written at the points Tspan into the array
                  YOutIn.
		  The parameters are:
		  Complex *Yout,  the array where the output should be
		                        stored in (columnwise storage).
		  finteger LdYout,      the leading dimension of Yout.
		  Real *Tspan, The array of the times for the dense
                                        output.
		  finteger NTspan,      the length of Tspan.

   AllFile:       The result vectors of all steps calculated by Exp4(Ex)
                  is written into a file.
		  The Parameters are:
		  FILE *file,           the file for writing the output.
		  (first number holds the time, and the rest of the line holds
		  the Y-values. ASCII 16 digits per value. Can be read by
		  matlab).

   DenseFile:     The result of the Dense output is written into 'FileIn',
                  same Format as above.
		  The Parameters are:
		  FILE *file,           the file for writing the output.
		  Real *Tspan, The array of the times for the dense
                                        output.
		  finteger NTspan,      the length of Tspan.


   EndFile:       The result of the last step is written into 'FileIn',
                  same Format as above.
		  The Parameters are:
		  FILE *file,           the file for writing the output.

   Solout:        The function SoloutFunctionIn is called after each step.
		  The Parameters are:
		  [SDCZ]SolTrait,       the function which is called after
		                        each step. */

extern int SExp4GetDenseSolution(fsingle *output,finteger incout,fsingle tout);
extern int DExp4GetDenseSolution(fdouble *output,finteger incout,fdouble tout);
extern int CExp4GetDenseSolution(fsinglecomplex *output,finteger incout,
      fsingle tout);
extern int ZExp4GetDenseSolution(fdoublecomplex *output,finteger incout,
      fdouble tout);
/* If you use the Solout function, you can obtain a dense output solution
   at the point tout by this function. This output is written into the
   array output with increment incout. Exp4GetDenseSolution does the same
   as the built in dense output routine. This gives you more freedom in
   the implementation of the solout function.
   Exp4GetDenseSolution returns 0 if tout is in the actual interval,
   else it returns 1. */

typedef enum { NormMax, Norm2 } Exp4NormType;
/* Enumeration Type for use with Exp4SetNormType, see below */

extern void SExp4SetNormType(Exp4NormType NormIn);
extern void DExp4SetNormType(Exp4NormType NormIn);
extern void CExp4SetNormType(Exp4NormType NormIn);
extern void ZExp4SetNormType(Exp4NormType NormIn);
/* You can change the norm which is used for error estimation.
   Default is the 2-norm. You can choose between the two values from
   above: NormMax is the infinity norm and Norm2 is the 2-norm. */

extern void SExp4ForceTime(fsingle TForceIn);
extern void DExp4ForceTime(fdouble TForceIn);
extern void CExp4ForceTime(fsingle TForceIn);
extern void ZExp4ForceTime(fdouble TForceIn);
/* With this function you force Exp4(Ex) to calculate an integration step
   at time TForceIn. That is if Dense output is accurate enough, you
   can call this function from the Solout function. Exp4 will then
   call the Solout function with solution y_new at the needed time.
   This is usually _NOT_ the next call of Solout, Solout is called
   after each step! */

extern fsingle *SExp4GetInternalStep(int i);
extern fdouble *DExp4GetInternalStep(int i);
extern fsinglecomplex *CExp4GetInternalStep(int i);
extern fdoublecomplex *ZExp4GetInternalStep(int i);
/* Returns a pointer to the i-th internal step */

extern fsingle *SExp4GetY();
extern fdouble *DExp4GetY();
extern fsinglecomplex *CExp4GetY();
extern fdoublecomplex *ZExp4GetY();
/* returns a pointer to the actual y value */

extern void SExp4EnableReuse(Exp4Boolean Allow);
extern void DExp4EnableReuse(Exp4Boolean Allow);
extern void CExp4EnableReuse(Exp4Boolean Allow);
extern void ZExp4EnableReuse(Exp4Boolean Allow);
/* If Allow is set to Exp4True (this is the default) Exp4 uses some
   reuse strategies (Set DEBUG to 2 and watch the screen output ;-) ).
   This strategy depends on a global continuous Jacobian. If your
   Jacobian is only local continuous it may be better not to use this
   reuse strategy, then set allow to Exp4False. */

extern void SExp4SetInitialStepsize(fsingle HInitIn);
extern void DExp4SetInitialStepsize(fdouble HInitIn);
extern void CExp4SetInitialStepsize(fsingle HInitIn);
extern void ZExp4SetInitialStepsize(fdouble HInitIn);
/* You can set the initial step size here. */

extern void SExp4SetJacobianLifeTime(int NSteps);
extern void DExp4SetJacobianLifeTime(int NSteps);
extern void CExp4SetJacobianLifeTime(int NSteps);
extern void ZExp4SetJacobianLifeTime(int NSteps);
/* You can set the maximum steps for a Jacobian to live */


#define TEST_SEQUENCE 1,2,3,4,6,8,11,15,20,27,36,46,57,70,85,100

/* defines the dimensions of the krylov space made in PhiKrylov
   for which the residuals (<=> error estimation) are checked */


extern void SExp4SetCGSDisableGuessFcn (Exp4Boolean Val);
extern void DExp4SetCGSDisableGuessFcn (Exp4Boolean Val);
extern void CExp4SetCGSDisableGuessFcn (Exp4Boolean Val);
extern void ZExp4SetCGSDisableGuessFcn (Exp4Boolean Val);
/* This switch is for internal use only! It could be used
 * to disable the fancy CGS strategies.
 */

extern void SExp4SetDisableCaching (Exp4Boolean Val);
extern void DExp4SetDisableCaching (Exp4Boolean Val);
extern void CExp4SetDisableCaching (Exp4Boolean Val);
extern void ZExp4SetDisableCaching (Exp4Boolean Val);
/* This switch is for internal use only! It could be used
 * to disable the chaching strategies for CGS and Lanczos.
 */

#ifdef  __cplusplus
}
#endif

#endif
