/*************************************************************************/
/* Copyright (C) 1995-1997 Marlis Hochbruck, Christian Lubich, Hubert    */
/* Selhofer. All rights reserved.                                        */
/*                                                                       */
/* C code written by Mathias Froehlich.                                  */
/*                                                                       */
/* This code is part of a copyrighted package. For details, see the file */
/* "COPYING" in the top-level directory.                                 */
/*************************************************************************/

/*
 This code is part of EXP4 and EXP4EX.

 This function computes (guesses) a starting step size for a RK-type
 method of order "ord".

 This strategy costs 2 function evaluations, but "f0" = f(t0,y0)
 can be reused in the algorithm.

 Reference: Hairer, Norsett, Wanner: "Solving Ordinary Differential
 Equations I. Nonstiff Problems", Springer 1992. (page 169ff)

 (C)  Copyright (c) 1-12-1996 by Marlis Hochbruck & Hubert Selhofer
      Marlis Hochbruck (marlis@na.uni-tuebingen.de)
      Hubert Selhofer (hubert@na.uni-tuebingen.de)

 If you are interested in the further development of the code
 and a proper maintenance, then please send wishes, comments,
 experiences, suggestions and bug reports to the addresses above!
 Thank you!

 C code written by Mathias Froehlich

 */

#include <math.h>
#include "ftypes.h"
#include "fblas.h"
#include "exp4.h"
#include "prec.h"
#include "work.h"
#include "errors.h"

#define LOCAL_DEBUG_OFFSET 4
#include "debug.h"

#include "scnorm.h"
#include "pfuns.h"

#include "h_init.h"


static RealCoef mone = RealCoefConst(-1.0);

static Real HInit = 0.0;

static finteger inc1=1;

void PREC(Exp4SetInitialStepsize) (Real HInitIn)
{
  HInit = HInitIn;
}

Real PREC(Exp4GetInitialStepsize) (WorkSpace Work,
      Complex *f0,finteger neq,Complex *y0,int ord)
/*
 * Exp4GetInitialStepsize approaches the first step size by calculating an
 * euler step.
 * Exp4GetInitalStepsize returns that approach.
 *
 * work      A pointer to the workspace for h_init.
 *           h_init requires neq*(2*sizeof(Complex)+sizeof(Real))
 *           bytes workspace.
 *
 * f0        A pointer to the function value at start point.
 *
 * neq       The number of equations, that is the dimension of im(RHSFunc).
 *
 * y0        A pointer to the start value y0. unchanged on exit.
 *
 * ord       The order of the method.
 */
{
  Real h,h0;
  RealCoef help;
  Complex *y1,*f1;
  Real md,d0=0.0,d1=0.0,d2;

  /* Quick return if stepsize is given.
     The RHS evaluation above is needed int the method later. */
  if (HInit != 0.0)
  {
    Exp4Message1((DEBUG_FILE,"Taking the given initial stepsize h=%e\n",
	  HInit));
    return HInit;
  }

  /* memory management */
  DistributeWorkspace(y1,neq*sizeof(Complex),Work);
  DistributeWorkspace(f1,neq*sizeof(Complex),Work);

  d0 = PREC(Exp4ScaledNorm) (y0);
  d1 = PREC(Exp4ScaledNorm) (f0);

  if ((d0>1e-5) && (d1>1e-5))
    h0 = 0.01*(d0/d1);
  else
    h0 = 1e-6;

  /* One Euler-step */
  pcopy (&neq,y0,&inc1,y1,&inc1);
  SetRealCoefVal(help,h0);
  praxpy (&neq,&help,f0,&inc1,y1,&inc1);

  /* A second guess */

  /*  f1 = feval(RHSFunc,tspan(1)+h0,y1); */
  PREC(Exp4EvalRhs) (neq,f1,y1);

  /* subtract f0 from f1 */
  praxpy (&neq,&mone,f0,&inc1,f1,&inc1);

  d2 = PREC(Exp4ScaledNormTmpScale) (f1,y1)/h0;

  md = max(d1,d2);
  if (md<1e-15)
    h = max(1e-6,h0*1e-3);
  else
    h = pow(1.0/(100.0*md),1.0/((fdouble)(ord+1)));

  h = min(100.0*h0,h);

  Exp4Message1((DEBUG_FILE,"Making an approach to the first step size h=%e\n",
	h));

  return h;
}
