/*************************************************************************/
/* Copyright (C) 1995-1997 Marlis Hochbruck, Christian Lubich, Hubert    */
/* Selhofer. All rights reserved.                                        */
/*                                                                       */
/* C code written by Mathias Froehlich.                                  */
/*                                                                       */
/* This code is part of a copyrighted package. For details, see the file */
/* "COPYING" in the top-level directory.                                 */
/*************************************************************************/

/* Header which defines some Macros important for uising Krylov spaces */

#ifndef EXP4_KRYLOV_H
#define EXP4_KRYLOV_H

#include "work.h"

#include "arnoldi.h"
#include "lanczos.h"
#include "cgs.h"

/* the type of the Krylov space which is represented with this handle */
typedef enum {
  Arn,
  Lan,
  LanSwitch,
  Cgs,
  CgsSwitch
} Exp4KrylovType;

/* argument to the initialization routine, indicates if the krylov space 
 * will be built or if there is only a linear combination to do */
typedef enum {
  Build,
  ReBuild
} Exp4KrylovInitType;

/* indicator which KrylovSub struct to use */
typedef enum {
  Primary = 0,
  Secondary = 1
} Exp4KrylovSwitchType;

/* a pointer to Krylov space type specific data */
typedef union {
  ArnoldiPrivat Arnoldi;
  LanczosPrivat Lanczos;
  CgsPrivat Cgs;
} KrylovPrivat;

/* Amount of memory for building the linear combination coeficients */
typedef size_t (* PhiZSizeWorkFunType) (finteger);

/* calculates the coeficients for the linear combination */
typedef void (* PhiZFunType) (WorkSpace,Real,finteger,Complex *,finteger,
      Real,finteger,finteger,Complex *,finteger);

/* Amount of memory for building the Krylov base */
typedef void (* KrylovSizeWorkFunType) (struct Krylov *,size_t *,size_t *);

/* initializes the start vectors and some other stuff */
typedef WorkSpace (* InitKryFunType) (WorkSpace,Complex *,Exp4KrylovInitType,
      struct Krylov *);

/* get H up to the size Guess, returns the really calcuated size of H */
typedef finteger (* KrylovGetHGuessFunType) (WorkSpace,Complex *,
      finteger,struct Krylov *);

/* increases the size of the base, returns a boolean wether the krylov
 * process should continue despite of good residuals */
typedef void (* KrylovIncreaseDimensionFunType) (WorkSpace,Real*,int*,
      Complex *,finteger,Complex *,finteger,finteger,Complex *,
      finteger,struct Krylov *);

/* calculates a linear combination of the base vectors */
typedef void (* KrylovCombineLinearFunType) (Complex *,finteger,Complex *,
      struct Krylov *,Complex *,finteger,finteger,int,int);


typedef struct SubKrylov {
  /* and the maximum Krylov dimension to calculate */
  finteger MaxSize;
  /* the number of superdiagonals of the matrix H */
  finteger bandwith;
  /* pointer to Krylov space type specific data */
  KrylovPrivat Privat;
  /* Amount of memory for building the Krylov base */
  KrylovSizeWorkFunType KrylovSizeWorkFun;
  /* initializes the start vectors and some other stuff */
  InitKryFunType InitKryFun;
  /* get H up to the size Guess, returns the really calcuated size of H */
  KrylovGetHGuessFunType KrylovGetHGuessFun;
  /* increases the size of the base, returns a boolean wether the krylov
   * process should continue despite of good residuals */
  KrylovIncreaseDimensionFunType KrylovIncreaseDimensionFun;
  /* calculates a linear combination of the base vectors */
  KrylovCombineLinearFunType KrylovCombineLinearFun;
} SubKrylov;

/* the Krylov space handle itself */
typedef struct Krylov {
  /* the type of the Krylov space */
  Exp4KrylovType Type;
  /* the leading dimenson of the lower dimensional stuff */
  finteger ldsmall;
  /* if krylov space can be rebuilt this is set to 1 */
  int CanRebuild;

  int LowStep;
  int GoodStep;
  double OptStep;
  double Exponent;
  int *Index;
  int IndexLen;

  /* the Workspace size which should be reserved for getting z
   * (as returned by PhiZSizeWorkFun) */
  size_t ReserveSize;
  /* Amount of memory for building the linear combination coeficients */
  PhiZSizeWorkFunType PhiZSizeWorkFun;
  /* calculates the coeficients for the linear combination */
  PhiZFunType PhiZFun;
  /* the length of each vector, the leading dimension of the base matrix  */
  finteger neq;
  /* the actual size of the Krylov space, for building a new space it
   * has to be zero on entry to PhiKry */
  finteger size;
  /* the dimension of the krylov space which was enough for convergence */
  finteger Konvergence;
  /* the matrix H itself */
  Complex *H;
  /* the norm of the start vector */
  Real Normv;
  /* holds the active Index */
  Exp4KrylovSwitchType Active;
  /* holds two possible sets of data for switching */
  SubKrylov KrylovData[2];
} krylov;



/* function to return the amount of memory for adding a krylov dimension */
#define KrylovSizeWork(Data,Shared,Resident) \
((Data)->KrylovData[KrylovActive(Data)].KrylovSizeWorkFun) \
(Data,Shared,Resident)

/* function to return the amount of memory for adding a krylov dimension */
#define KrylovReserveSize(Data)      ((Data)->ReserveSize)

/* function to initialize the start vectors */
#define InitKry(Data,Work,Start,Type) \
((Data)->KrylovData[KrylovActive(Data)].InitKryFun) (Work,Start,Type,Data)

/* function to increase the Krylov dimension */
#define KrylovIncreaseDimension(Work,RNrm,Accept,x,ldx,z,ldz,fac_t,Start, \
      UpTo,Data) \
((Data)->KrylovData[KrylovActive(Data)].KrylovIncreaseDimensionFun) \
(Work,RNrm,Accept,x,ldx,z,ldz,fac_t,Start,UpTo,Data)

#define KrylovGetHGuess(Work,Start,Guess,Data) \
((Data)->KrylovData[KrylovActive(Data)].KrylovGetHGuessFun) \
(Work,Start,Guess,Data);

#define KrylovHasHGuess(Data) \
((Data)->KrylovData[KrylovActive(Data)].KrylovGetHGuessFun != NULL)


/* function to calculate a linear combination of the Krylov base */
#define KrylovCombineLinear(Work,ldx,Start,Data,zz,ldz,fac_t,HaveOldV,FrcReb) \
     ((Data)->KrylovData[KrylovActive(Data)].KrylovCombineLinearFun) \
        (Work,ldx,Start,Data,zz,ldz,fac_t,HaveOldV,FrcReb)

/* function to return the amount of memory for building the linear
 * combination coeficients */
#define PhiZSizeWork(Data,MaxDim) ((Data)->PhiZSizeWorkFun) (MaxDim)

/* function to calculate the coeficients for the linear combination */
#define PhiZ(Data,Work,Alpha,nZColumns,Z,ldZ,Beta,k,Band,H,ldH) \
     ((Data)->PhiZFun) (Work,Alpha,nZColumns,Z,ldZ,Beta,k,Band,H,ldH)

/* accesses the type of the krylov process */
#define KrylovType(Data)        ((Data)->Type)

/* is true if the Krylov space can be rebuilt */
#define KrylovCanRebuild(Data)  ((Data)->CanRebuild)

/* accesses the norm of the start vector for the krylov process */
#define KrylovNormv(Data)       ((Data)->Normv)

/* accesses the number of equations of the krylov process */
#define KrylovNeq(Data)         ((Data)->neq)

/* accesses the leading dimension of V and W */
#define KrylovLdBig(Data)       ((Data)->neq)

/* accesses the leading dimension of H */
#define KrylovLdSmall(Data)     ((Data)->ldsmall)

/* accesses the pointer to H */
#define KrylovH(Data)           ((Data)->H)

/* accesses the number of superdiagonals of H */
#define KrylovHBandWith(Data) \
     ((Data)->KrylovData[KrylovActive(Data)].bandwith)

/* accesses the i, j - th entry of H */
#define KrylovHent(Data,i,j)    (*((Data)->H+(i)+(j)*KrylovLdSmall(Data)))

/* accesses the size of the krylov space */
#define KrylovSize(Data)        ((Data)->size)

/* accesses the size of the krylov space */
#define KrylovKonvergence(Data) ((Data)->Konvergence)

/* clears the krylov space pointed to by Data */
#define ClearKrylovSpace(Data)  ((Data)->size) = 0

#define KrylovActive(Data)      ((Data)->Active)

#define KrylovActivePrimary(Data) \
     (KrylovActive(Data) = Primary)

#define KrylovIsPrimary(Data) \
     (KrylovActive(Data) == Primary)

#define KrylovActiveSecondary(Data) \
     (KrylovActive(Data) = Secondary)

#define KrylovIsSecondary(Data) \
     (KrylovActive(Data) == Secondary)

#define KrylovHasSecondary(Data) \
     (((Data)->Type == LanSwitch) || ((Data)->Type == CgsSwitch))

/* accesses the size of the krylov space */
#define KrylovMaxSize(Data) \
     ((Data)->KrylovData[KrylovActive(Data)].MaxSize)

#define KrylovPrimaryMaxSize(Data) \
     ((Data)->KrylovData[Primary].MaxSize)

#define KrylovLowStep(Data)     ((Data)->LowStep)

#define KrylovGoodStep(Data)    ((Data)->GoodStep)

#define KrylovOptStep(Data)     ((Data)->OptStep)

#define KrylovExponent(Data)    ((Data)->Exponent)

#define KrylovIndex(Data)       ((Data)->Index)
#define KrylovIndexLen(Data)    ((Data)->IndexLen)

#define KrylovPrivat(Data)      ((Data)->KrylovData[KrylovActive(Data)].Privat)

/* prototypes of functions defined in krylov.c */
#ifdef  __cplusplus
extern "C" {
#endif

extern void PREC(Exp4GetKrylovSpaceHandles) (krylov **Kry1,krylov **Kry2,
      krylov **Kry3,finteger neq);

extern struct Krylov *PREC(Exp4GetKrylovHandle) (Exp4KrylovType Type,
      finteger MaxKrylovDim,finteger MaxBlockSize,finteger neq,int CanRebuild,
      PhiZFunType PhiZFun,PhiZSizeWorkFunType PhiZSizeWorkFun);

#ifdef  __cplusplus
}
#endif

#endif /* defined EXP4_KRYLOV_H */
