/*************************************************************************/
/* Copyright (C) 1995-1997 Marlis Hochbruck, Christian Lubich, Hubert    */
/* Selhofer. All rights reserved.                                        */
/*                                                                       */
/* C code written by Mathias Froehlich.                                  */
/*                                                                       */
/* This code is part of a copyrighted package. For details, see the file */
/* "COPYING" in the top-level directory.                                 */
/*************************************************************************/
/*
 * This code is part of EXP4.
 *
 * This module builds and evaluates the krylov space by Lanczos recursions
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <math.h>
#include "ftypes.h"
#include "fblas.h"
#include "flapack.h"
#include "exp4.h"
#include "prec.h"
#include "errors.h"

/* #define LOCAL_DEBUG_LEVEL 20 */
#define LOCAL_DEBUG_OFFSET 6
#include "debug.h"

#include "pfuns.h"
#include "work.h"
#include "krylov.h"
#include "scnorm.h"
#include "krystep.h"

#include "lanczos.h"

/* #define USE_SVD */
#define USE_QR_DECOMPOSITION

#define COEF_NORM_TOLERANCE 1.1

#define MAX_LANCZOS_TRIES  2

/* #define ADDITIONAL_ORTHO */
/* #define ADDITIONAL_ORTHO_II */

/* #define ARNOLDI_ORTHO     */

#define ARNOLDI_ORTHO_II    
#define ARNOLDI_BAND 1

/* #define STORE_FULL */

/* #define CHECK_D */
/* #define DIRECT_D */
/* #define CHECK_RHS */
/* #define DIRECT_RHS */

#ifdef SINGLE_PRECISION
#define PPRE "s"
#else
#ifdef DOUBLE_PRECISION
#define PPRE "d"
#else
#ifdef SINGLE_COMPLEX
#define PPRE "c"
#else
#ifdef DOUBLE_COMPLEX
#define PPRE "z"
#else
#error "Incorrect Precision set"
#endif
#endif
#endif
#endif

static finteger inc0 = 0;
static finteger inc1 = 1;

static Complex wzero = ComplexConst(0.0);
static Complex wone  = ComplexConst(1.0);


/* structure to hold a decomposition of D */
typedef struct {
  finteger n;
#ifdef USE_SVD
  Complex *U,*VT;
  Real *S;
#else
  Complex *Data;
#ifdef USE_QR_DECOMPOSITION
  Complex *Tau;
#endif
  finteger *Pivot;
#endif
} Decomposition;


/* the lower bound for the singular values */
static Real SingularBound;


#ifdef STORE_FULL
extern void PREC(StoreLanczosVect) (krylov *Kry,Complex *V,
      Complex *W,finteger k);
extern void PREC(PrintVW) (krylov *Kry);
extern void PREC(CompareVW) (krylov *Kry);
extern void PREC(testAV_VH) (krylov *Kry,finteger k,int norm);
extern void PREC(BiOrth) (krylov *Kry,finteger k,double thresh);
extern void PREC(DHNorm) (krylov *Kry,finteger k);
#endif
extern void flaprintf(FILE *,char *,...);


/* returns the size of workspace required for Exp4AddKrylovDimension */
static void PREC(Exp4SizeLanczosWork) (krylov *Kry,size_t *Shared,
      size_t *Resident);

/* sets up the base and writes the norm of the start vector */
static WorkSpace PREC(LanczosInit) (WorkSpace Work,Complex *Start,
      Exp4KrylovInitType What,krylov *Kry);

/* increases the krylov space */
static void PREC(Exp4IncreaseDimensionLanczos) (WorkSpace Work,
      Real *ResiualNorm,int *AcceptKonvergence,
      Complex *x,finteger ldx,Complex *zz,finteger ldz,
      finteger fac_t,Complex *Start,finteger UpTo,krylov *Kry);

/* rebuilds the V base and calculate the linear combination */
static void PREC(Exp4KrylovCombineLinearLanczos) (Complex *x,
      finteger ldx,Complex *Start,krylov *Kry,
      Complex *zz,finteger ldz,finteger fac_t,int HaveOldV,int dummy);




/* calculates the coeficients needed for closing a block */
static Real PREC(GetAlpha) (WorkSpace Work,Complex *AlphaV,
      Complex *AlphaW,krylov *Kry,Real *NormV,
      Real *NormW,Complex *D__,finteger LdD,
      Complex *HV__,finteger LdHV,Complex *HW__,finteger LdHW,
      Decomposition *Dinv,finteger k);

/* calculates the coeficients for the last block */
static Real PREC(GetBeta) (Complex *BetaV,Complex *BetaW,
      Complex *D__,finteger LdD,Real *NormV,Real *NormW,
      krylov *Kry,finteger k);

/* decomposites the matrix m and writes the decomposition into Minv */
static Real PREC(DecompositeMatrix) (WorkSpace Work,
      Decomposition *Minv,finteger k,Complex *m,finteger ldm);

/* solves the linear equation Minv*b, overwrites b */
static finteger PREC (ApplyDecomposition) (WorkSpace Work,
      Decomposition *Minv,int CmplxTransp,Complex *b);

/* apply's the coeficients AlphaV and AlphaW to the bases V,W */
static void PREC(OrthogonalizeToCurrentBlock) (krylov *Kry,
      Complex *AlphaV,Complex *AlphaW,
      Complex *HV__,finteger LdHV,Complex *HW__,finteger LdHW,
      finteger k,finteger regular);

/* apply's the coeficients BetaV and BetaW to the bases V,W */
static void PREC(OrthogonalizeToLastBlock) (krylov *Kry,Complex BetaV,
      Complex BetaW,finteger k);

/* calculates the temporary vector to close a block */
static void PREC(CalculateTemporary) (WorkSpace Work,krylov *Kry,
      Decomposition *Dinv,finteger k);

/* calculates a new column and line of D */
static void PREC(UpdateD) (krylov *Kry,Complex *D__,finteger LdD,
      Real *NormV,Real *NormW,
      Complex *HV__,finteger LdHV,Complex *HW__,finteger LdHW,
      finteger k,int V0_EQ_W0);

/* normalizes the new vectors in V, W and store their norms in NormV, NormW */
static finteger PREC(Normalize) (Real *NormV,Real *NormW,
      krylov *Kry,finteger k);

/* returns the best blocksize found up to now */
static finteger PREC(BestBlockSize) (Real *CoefNorm,finteger k);


/* returns a Lanczos Krylov 'Handle' */
void PREC(Exp4FillHandleWithLanczos) (struct Krylov *Kry,finteger MaxKrylovDim,
      finteger MaxBlockSize,finteger neq,int HaveOwnH)
{
  Kry->KrylovData[KrylovActive(Kry)].KrylovSizeWorkFun =
    & PREC(Exp4SizeLanczosWork);
  Kry->KrylovData[KrylovActive(Kry)].InitKryFun =
    & PREC(LanczosInit);
  Kry->KrylovData[KrylovActive(Kry)].KrylovGetHGuessFun = NULL;
  Kry->KrylovData[KrylovActive(Kry)].KrylovIncreaseDimensionFun =
    & PREC(Exp4IncreaseDimensionLanczos);
  Kry->KrylovData[KrylovActive(Kry)].KrylovCombineLinearFun =
    & PREC(Exp4KrylovCombineLinearLanczos);

  SingularBound = pow(plamch("Precision"),4.0/5.0);

  KrylovMaxSize(Kry) = MaxKrylovDim;
  KrylovHBandWith(Kry) = MaxBlockSize*2 - 1;
  LanczosMaxBlockSize(Kry) = MaxBlockSize;
  LanczosWStride(Kry) = MaxBlockSize + 2;
  LanczosVStride(Kry) = MaxBlockSize + 2;
  
  if (HaveOwnH)
  {
    WorkSpace Work;
    size_t BetaWork = sizeof(Complex)*(KrylovMaxSize(Kry)+1);
    size_t RegWork = sizeof(finteger)*(KrylovMaxSize(Kry)+1);

    size_t Size = 2*GetWorkSpaceSize(BetaWork);
    Size += GetWorkSpaceSize(RegWork);

    Work = Exp4GetWorkspace(Size);

    DistributeWorkspace(LanczosBetaTemp(Kry),BetaWork,Work);
    DistributeWorkspace(LanczosBetaMult(Kry),BetaWork,Work);
    DistributeWorkspace(LanczosRegularIndexData(Kry),RegWork,Work);
  }

  Exp4Message1((DEBUG_FILE,"Created Lanczos Handle: Maximum Dimension = %d,"
	" Maximum BlockSize = %d\n",KrylovMaxSize(Kry),
	LanczosMaxBlockSize(Kry)));
}


/* returns the size of workspace required for Exp4AddKrylovDimension */
static void PREC(Exp4SizeLanczosWork)  (krylov *Kry,size_t *Shared,
      size_t *Resident)
{
  finteger LdBlock = LanczosMaxBlockSize(Kry)+1;  
  /* workspace for storing D */
  (*Shared) = GetWorkSpaceSize(sizeof(Complex)*LdBlock*LdBlock);

  /* is needed for D, Dinv, AlphaV, AlphaW, HW */
#ifdef USE_SVD
  (*Shared) *= 6;
#else
  (*Shared) *= 5;
#endif

  /* Space for BetaV, BetaW */
  (*Shared) += 2*GetWorkSpaceSize(sizeof(Complex)*LdBlock);

#if defined(USE_QR_DECOMPOSITION) || defined(USE_SVD)
  /* Space for Dinv.Tau or DInv.S in case of singular value decomposition */
  (*Shared) += GetWorkSpaceSize(sizeof(Complex)*LdBlock);
#endif

  /* Space for NormV, NormW, CoefNorm */
  (*Shared) += 3*GetWorkSpaceSize(sizeof(Real)*LdBlock);

  /* Space for Dinv.Perm */
  (*Shared) += GetWorkSpaceSize(sizeof(finteger)*LdBlock);

  /* Space for Decomposite Matrix and ApplyDecomposition */
#ifdef USE_SVD
#if (defined(DOUBLE_PRECISION) || defined(SINGLE_PRECISION))
  (*Shared) += GetWorkSpaceSize(sizeof(Complex)*LdBlock*5);
#else
  (*Shared) += GetWorkSpaceSize(sizeof(Real)*LdBlock*5);
  (*Shared) += GetWorkSpaceSize(sizeof(Complex)*LdBlock*3);
#endif
#else
  (*Shared) += GetWorkSpaceSize(sizeof(Complex)*LdBlock*3);
#endif

  /* minimal requirement for V and W */
  (*Resident) = 2*GetWorkSpaceSize(sizeof(Complex)*LanczosWStride(Kry)*
	KrylovNeq(Kry));

  if (!KrylovCanRebuild(Kry))
  {
    size_t HWork = sizeof(Complex)*KrylovLdSmall(Kry)*KrylovMaxSize(Kry);
    size_t BetaWork = sizeof(Complex)*(KrylovMaxSize(Kry)+1);
    size_t RegWork = sizeof(finteger)*(KrylovMaxSize(Kry)+1);

    (*Resident) += GetWorkSpaceSize(HWork);
    (*Resident) += 2*GetWorkSpaceSize(BetaWork);
    (*Resident) += GetWorkSpaceSize(RegWork);
  }
}

/* defines for building the krylov space */
#define VN(I) (LanczosV(Kry,(KrylovSize(Kry)+(I))%LanczosVStride(Kry)))
    
#define WN(I) (LanczosW(Kry,(KrylovSize(Kry)+(I))%LanczosWStride(Kry)))

#define InvalidateVN(k)							      \
{									      \
  finteger __LastValidOld = KrylovPrivat(Kry).Lanczos.LastValidV;	      \
  finteger __LastValidNew = KrylovSize(Kry)+(k)+1;			      \
  KrylovPrivat(Kry).Lanczos.LastValidV = max(__LastValidOld,__LastValidNew);  \
  Exp4Message1((DEBUG_FILE,"Invalidating V(%d)\n",KrylovSize(Kry)+(k)));      \
} 

static WorkSpace PREC(LanczosInit) (WorkSpace Work,Complex *Start,
      Exp4KrylovInitType What,krylov *Kry)
{
  switch (What) {
  case Build:
    {
      if (!KrylovCanRebuild(Kry))
      {
	size_t HWork = sizeof(Complex)*KrylovLdSmall(Kry)*
	  KrylovMaxSize(Kry);
	size_t BetaWork = sizeof(Complex)*(KrylovMaxSize(Kry)+1);
	size_t RegWork = sizeof(finteger)*(KrylovMaxSize(Kry)+1);
	
	DistributeWorkspace(KrylovH(Kry),HWork,Work);
	DistributeWorkspace(LanczosBetaTemp(Kry),BetaWork,Work);
	DistributeWorkspace(LanczosBetaMult(Kry),BetaWork,Work);
	DistributeWorkspace(LanczosRegularIndexData(Kry),RegWork,Work);
      }

      {
	int NVec = (GetRemainingWorkSpaceSize(Work)-KrylovReserveSize(Kry))/
	  (sizeof(Complex)*KrylovNeq(Kry));
	size_t Len = sizeof(Complex)*NVec*KrylovNeq(Kry);

	Len = max(Len,2*(LanczosMaxBlockSize(Kry)+2)*
	      sizeof(Complex)*KrylovNeq(Kry));

	LanczosVStride(Kry) = max(NVec - LanczosWStride(Kry),
	      LanczosMaxBlockSize(Kry)+2);

	DistributeWorkspace(LanczosVData(Kry),Len,Work);
	LanczosWData(Kry)=LanczosVData(Kry)+LanczosVStride(Kry)*KrylovNeq(Kry);

	Exp4Message1((DEBUG_FILE,"Taking %d vectors for V caching\n",
	      (int)LanczosVStride(Kry)));
      }

      /* get the norm of the start vector */
/*       KrylovNormv(Kry) = pnrm2 (&KrylovNeq(Kry),Start,&inc1); */

      break;
    }
  case ReBuild:
    {
      if (!KrylovCanRebuild(Kry))
	ErrorReturn("Can not rebuild unknown krylov space!",
	      EXP4_ERR_TRY_TO_REBUILD_UNKNOWN_KRYLOV_SPACE);

      LanczosVStride(Kry) = LanczosWStride(Kry);
      {
	size_t VWorkLen = sizeof(Complex)*LanczosVStride(Kry)*
	  KrylovNeq(Kry);
	DistributeWorkspace(LanczosVData(Kry),VWorkLen,Work);
	DistributeWorkspace(LanczosWData(Kry),VWorkLen,Work);
      }

      break;
    }
  }

  return Work;
}

static void PREC(Exp4IncreaseDimensionLanczos) (WorkSpace Work,
      Real *ResiualNorm,int *AcceptKonvergence,
      Complex *x,finteger ldx,Complex *zz,finteger ldz,
      finteger fac_t,Complex *Start,finteger UpTo,krylov *Kry)
{
#define D(i,j) D__[(i)+(j)*LdBlock]
#define HW(i,j) HW__[(i)+(j)*LdBlock]
#define AlphaV(k) (AlphaV__+((k)-1)*(LdBlock))
#define AlphaW(k) (AlphaW__+((k)-1)*(LdBlock))
  
  /* indicates if an exact breakdown occures */
  int ExactBreakdown = 0;
  /* indicates if the start vector for the W base is equal to the
   * V start vector */
  int W0_EQ_V0 = 1;
  /* the diagonal entries of W'*V */
  Complex *D__;
  /* the 'local' entries of the block tridiagonal matrix H of the
   * Krylov space W to A'. Is needed to calculate the D entries recursive. */
  Complex *HW__;
  /* the leading dimension for all small matrices */
  finteger LdBlock = LanczosMaxBlockSize(Kry)+1;

  /* the inner coeficients needed for closing a block */
  Complex *AlphaV__;
  Complex *AlphaW__;
  /* the coeficients of the last block */
  Complex *BetaV;
  Complex *BetaW;
  /* the norms of the V and W vectors */
  Real *NormV;
  Real *NormW;
  /* the norms of the coeficient vectors in each block */
  Real *CoefNorm;
  /* the inverse of the 'local' matrix D */
  Decomposition Dinv;

  /* workspace distribution */
  DistributeWorkspace(D__,sizeof(Complex)*LdBlock*LdBlock,Work);
#ifdef USE_SVD
  DistributeWorkspace(Dinv.U,sizeof(Complex)*LdBlock*LdBlock,Work);
  DistributeWorkspace(Dinv.VT,sizeof(Complex)*LdBlock*LdBlock,Work);
  DistributeWorkspace(Dinv.S,sizeof(Complex)*LdBlock,Work);
#else
  DistributeWorkspace(Dinv.Data,sizeof(Complex)*LdBlock*LdBlock,Work);
#ifdef USE_QR_DECOMPOSITION
  DistributeWorkspace(Dinv.Tau,sizeof(Complex)*LdBlock,Work);
#endif
  DistributeWorkspace(Dinv.Pivot,sizeof(finteger)*LdBlock,Work);
#endif
  DistributeWorkspace(HW__,sizeof(Complex)*LdBlock*LdBlock,Work);

  DistributeWorkspace(AlphaV__,sizeof(Complex)*LdBlock*LdBlock,Work);
  DistributeWorkspace(AlphaW__,sizeof(Complex)*LdBlock*LdBlock,Work);

  DistributeWorkspace(BetaV,sizeof(Complex)*LdBlock,Work);
  DistributeWorkspace(BetaW,sizeof(Complex)*LdBlock,Work);

  DistributeWorkspace(NormV,sizeof(Real)*LdBlock,Work);
  DistributeWorkspace(NormW,sizeof(Real)*LdBlock,Work);
  DistributeWorkspace(CoefNorm,sizeof(Real)*LdBlock,Work);

  /* calculate new base vectors up to the dimension UpTo */
  while (KrylovSize(Kry) < UpTo)
  {
    /* flag which indicates if this block is regular or not */
    int regular = 0;
    /* the actual block size */
    finteger k = 0;

    /* if this point is reached from an exact breakdown, restart the
     * process with a random start vector for the W base */
    if (ExactBreakdown)
    {
      /* reset the krylov space size */
      KrylovSize(Kry) = 0;
      /* markthe first W vector different from the first V */
      W0_EQ_V0 = 0;
      /* reset the flag for exact breakdown */
      ExactBreakdown = 0;
      /* count this breakdown */
      LanczosIncrementExactBreakdownCount(Kry);
#if (DEBUG >= 1)
      fprintf(stdout," Lanczos Process had an unrecoverable breakdown: "
	    "restarting\n");
#endif
    }

    if (KrylovSize(Kry) == 0)
    {
      /* normalize the first vector in the Krylov base */
      pcopy (&KrylovNeq(Kry),Start,&inc1,VN(0),&inc1);
      prscl (&KrylovNeq(Kry),&KrylovNormv(Kry),VN(0),&inc1);
      if (W0_EQ_V0)
      {
	Exp4Message2((DEBUG_FILE,"Using V(0) for W(0)\n"));
	/* take the first W vector equal to the first V vector */
	pcopy (&KrylovNeq(Kry),VN(0),&inc1,WN(0),&inc1);
	LanczosClearExactBreakdownCount(Kry);
      }
      else
      {
	Real norm;
	finteger idist = 2;
	static finteger iseed[4] = {324,567,435,775};
	Exp4Message2((DEBUG_FILE,"Using a random vector for W(0)\n"));
	/* take the first W to a random vector */
	plarnv (&idist,iseed,&KrylovNeq(Kry),WN(0));
	/* normalize W */
	norm = pnrm2 (&KrylovNeq(Kry),WN(0),&inc1);    
	prscl (&KrylovNeq(Kry),&norm,WN(0),&inc1);
      }
      
#ifdef STORE_FULL
      PREC(StoreLanczosVect) (Kry,VN(0),WN(0),0);
#endif
      
      ComplexSetReal(LanczosLastD(Kry),1.0);
      LanczosClearRegularIndexArray(Kry);
      LanczosLastValidV(Kry) = 0;
      LanczosNorm1BetaV(Kry) =  LanczosNorm1BetaW(Kry) = 1.0;
    }
    
    /* build a block if needed and as long as possible */
    while (!regular && (k < LanczosMaxBlockSize(Kry)) &&
	  ((KrylovSize(Kry)+k) < KrylovMaxSize(Kry)))
    {
      /* the smallest singlular value of D */
      Real Sing;
      /* increment the block size */
      k++;
     
      /* do the scaling of VK, WK and write entry in H */
      if (PREC(Normalize) (NormV,NormW,Kry,k-1))
      {
	if (LanczosExactBreakdownCount(Kry) >= MAX_LANCZOS_TRIES)
	  ErrorReturn("Found exact breakdown in Lanczos Process!\n"
		"Please consider taking Arnoldi!\n"
		"Abort ...\n",EXP4_ERR_EXACT_BREAKDOWN);
	ExactBreakdown = 1;
	break;
      }

      /* mark the vector V(n+1-VectorStride) as overwritten,
       * needed for rebuilding the base vectors */
      InvalidateVN(k-LanczosVStride(Kry));
      /* V(n+1) = Jacobian * V(n) */  
      PREC(Exp4Mv) (KrylovNeq(Kry),VN(k),VN(k-1),1);
      /* W(n+1) = conjugate_transpose(Jacobian) * W(n) */  
      PREC(Exp4vM) (KrylovNeq(Kry),WN(k),WN(k-1));

      /***********************************************************************/
      if ((KrylovSize(Kry) == 0) && (PREC(Exp4GetNormJacApproach) () == 0.0))
	PREC(Exp4SetNormJacApproach) (pnrm2(&KrylovNeq(Kry),VN(k),&inc1)),
	  PREC(Exp4SetNormJacApproach) (pnrm2(&KrylovNeq(Kry),WN(k),&inc1));
      /***********************************************************************/

      /* calculate the new entries in D */
      PREC(UpdateD) (Kry,&D(0,0),LdBlock,NormV,NormW,
	    &KrylovHent(Kry,KrylovSize(Kry),KrylovSize(Kry)),
	    KrylovLdSmall(Kry),&HW(0,0),LdBlock,k,W0_EQ_V0);

      /* calculate the coeficients for the last block */
      CoefNorm[k] = PREC(GetBeta) (BetaV+k,BetaW+k,&D(0,0),LdBlock,
	    NormV,NormW,Kry,k);

      /* apply these coeficients */
      PREC(OrthogonalizeToLastBlock) (Kry,BetaV[k],BetaW[k],k);
      
      /* calculate " D^-1 " */
      Sing = PREC(DecompositeMatrix) (Work,&Dinv,k,&D(0,0),LdBlock);

      /* calculate w'*A*v and write it into D(k-1,k),
       * UpdateD reuses this value to calculate D(k-1,k) */
      pdot (&D(k-1,k),&KrylovNeq(Kry),WN(k-1),&inc1,VN(k),&inc1);


      Exp4Message2((DEBUG_FILE,"Size = %d, k = %d\n",KrylovSize(Kry),k));
      Exp4Message3((DEBUG_FILE,"Smallest singular value of D: %e\n",Sing));
      Exp4Message3((DEBUG_FILE,"Coefficient norm of Beta: %e\n",CoefNorm[k]));

      /* if D is nonsingular test if the block is regular */
      if ((Sing > SingularBound) || (LanczosMaxBlockSize(Kry) == 1))
      {
	/* help variable */
	Real thelp;

	/* detect an exact breakdown in the classical Lanczos */
	if ((LanczosMaxBlockSize(Kry) == 1) && (Sing == 0.0))
	{
	  if (LanczosExactBreakdownCount(Kry) >= MAX_LANCZOS_TRIES)
	    ErrorReturn("Found exact breakdown in Lanczos Process!\n"
		  "Please consider taking Arnoldi!\n"
		  "Abort ...\n",EXP4_ERR_EXACT_BREAKDOWN);
	  ExactBreakdown = 1;
	  break;
	}

	/* calculate the coeficients AlphaV and AlphaW */
	thelp = PREC(GetAlpha) (Work,AlphaV(k),AlphaW(k),Kry,NormV,NormW,
	      &D(0,0),LdBlock,&KrylovHent(Kry,KrylovSize(Kry),KrylovSize(Kry)),
	      KrylovLdSmall(Kry),&HW(0,0),LdBlock,&Dinv,k);
	CoefNorm[k] = max(CoefNorm[k],thelp);

	Exp4Message3((DEBUG_FILE,"Coefficient norm of Alpha and Beta: %e\n",
	      CoefNorm[k]));
	Exp4Message3((DEBUG_FILE,"Approach to || A || = %e\n",
	      PREC(Exp4GetNormJacApproach) ()));
      
	/* check if current step is regular */
	if ((CoefNorm[k] < PREC(Exp4GetNormJacApproach) ()) ||
	      (LanczosMaxBlockSize(Kry) == 1))
	  regular = 1;
	else
	{
	  /* if we'v reached the biggest allowed blocksize, and it is the
	   * best k within this block accept it as regular */
	  if ((k == LanczosMaxBlockSize(Kry)) &&
		(PREC(BestBlockSize) (CoefNorm,k) == k))
	  {
	    PREC(Exp4SetNormJacApproach) (CoefNorm[k]);
	    regular = 1;
	  }
	  else
	    regular = 0;
	}
      }
      else
      {
	if (Sing <= SingularBound) CoefNorm[k] = -1;

	/* if we'v reached the biggest allowed blocksize, and it is the
	 * best k within this block accept it as regular */
	if ((k == LanczosMaxBlockSize(Kry)) &&
	      (PREC(BestBlockSize) (CoefNorm,k) == k))
	{
	  PREC(Exp4SetNormJacApproach) (CoefNorm[k]);
	  regular = 1;
	}
	else
	  regular = 0;
      }
      /* apply the AlphaV and AlphaW if regular */
      PREC(OrthogonalizeToCurrentBlock) (Kry,AlphaV(k),AlphaW(k),
	    &KrylovHent(Kry,KrylovSize(Kry),KrylovSize(Kry)),
	    KrylovLdSmall(Kry),&HW(0,0),LdBlock,k,regular);
    }
    if (ExactBreakdown) continue;

    if (!regular)
    {
      k = PREC(BestBlockSize) (CoefNorm,k);
      regular = 1;

      if (k == 0)
      {
#ifdef STORE_FULL
/*   PREC(PrintVW) (Kry); */
/*   PREC(BiOrth) (Kry,0,1e-100); */
#endif
        if (LanczosExactBreakdownCount(Kry) >= MAX_LANCZOS_TRIES)
	  ErrorReturn("Found exact breakdown in Lanczos Process!\n"
		"Please consider taking Arnoldi!\n"
		"Abort ...\n",EXP4_ERR_EXACT_BREAKDOWN);
	ExactBreakdown = 1;
	continue;
      }
	
      /* The decomposition is needed for calculating the temporary vector */
      PREC(DecompositeMatrix) (Work,&Dinv,k,&D(0,0),LdBlock);

      PREC(Exp4SetNormJacApproach) (CoefNorm[k]);

      /* V(n+1) = Jacobian * V(n) */  
      PREC(Exp4Mv) (KrylovNeq(Kry),VN(k),VN(k-1),1);
      /* W(n+1) = conjugate_transpose(Jacobian) * W(n) */  
      PREC(Exp4vM) (KrylovNeq(Kry),WN(k),WN(k-1));
      
      PREC(OrthogonalizeToLastBlock) (Kry,BetaV[k],BetaW[k],k);

      PREC(OrthogonalizeToCurrentBlock) (Kry,AlphaV(k),AlphaW(k),
	    &KrylovHent(Kry,KrylovSize(Kry),KrylovSize(Kry)),
	    KrylovLdSmall(Kry),&HW(0,0),LdBlock,k,regular);
    }

    if (k == 1)
    {
      LanczosLastD(Kry) = D(0,0);
      LanczosNorm1BetaV(Kry) =  LanczosNorm1BetaW(Kry) = 1.0;
    }
    else
    {
      PREC(CalculateTemporary) (Work,Kry,&Dinv,k);
      ComplexSetReal(LanczosLastD(Kry),1.0);
    }

    Exp4Message2((DEBUG_FILE,"Size = %d, Calculated block of size %d\n",
	  KrylovSize(Kry),k));

    KrylovSize(Kry) += k;
    LanczosSetRegularIndex(Kry,KrylovSize(Kry));

/* PREC(BiOrth) (Kry,0,1e-14); */
/* flaprintf(stdout,"V =\n%ladvt.1e\nW =%ladvt.1e\n",KrylovNeq(Kry),VN(0),1,KrylovNeq(Kry),WN(0),1); */
  }

/* PREC(testAV_VH) (Kry,0,1); */
/* PREC(BiOrth) (Kry,0,1e-16); */
/* flaprintf(stdout,"H =\n%la"PPRE"gee\n",KrylovSize(Kry)+1,KrylovSize(Kry), */
/*       KrylovH(Kry),KrylovLdSmall(Kry)); */
/* PREC(DHNorm) (Kry,0); */

  (*ResiualNorm) = PREC(Exp4ScaledNorm) (VN(0));
  (*AcceptKonvergence) = 1;

#undef D
#undef HW
#undef AlphaV
#undef AlphaW
}

static Real PREC(GetAlpha) (WorkSpace Work,Complex *AlphaV,
      Complex *AlphaW,krylov *Kry,Real *NormV,
      Real *NormW,Complex *D__,finteger LdD,
      Complex *HV__,finteger LdHV,Complex *HW__,finteger LdHW,
      Decomposition *Dinv,finteger k)
{
#define D(i,j) D__[(i)+(j)*LdD]
#define HV(i,j) HV__[(i)+(j)*LdHV]
#define HW(i,j) HW__[(i)+(j)*LdHW]

  Real thelp;
  Real CoefNorm;

  /* Alpha = w'*A*v / D(0,0), w'*A*v was written into D(k-1,k)
   * it can be reused for D(k-1,k), D(k,k-1). */
  AlphaV[k-1] = D(k-1,k);
  conj(AlphaW[k-1],AlphaV[k-1]);

  if (k == 1)
  {
    pdiv(AlphaV[0],AlphaV[0],D(0,0));
    conj(AlphaW[0],AlphaV[0]);
    CoefNorm = pabs(AlphaV[0]);
  }
  else
  { 
    finteger i;

#if (defined(CHECK_RHS) || defined(DIRECT_RHS))
    Complex AV[k],AW[k];
    for (i=0;i<k;i++)
    {
      pdot (&AV[i],&KrylovNeq(Kry),WN(i),&inc1,VN(k),&inc1);
      pdot (&AW[i],&KrylovNeq(Kry),VN(i),&inc1,WN(k),&inc1);
    }
#endif

    for (i=0;i<k-1;i++)
    {
      int j;
      rmul(AlphaV[i],NormW[i+1],D(i+1,k-1));
      for (j=0;j<=i;j++)
      {
	Complex whelp;
	conj(whelp,HW(j,i));
	mul(whelp,whelp,D(j,k-1));
	add(AlphaV[i],AlphaV[i],whelp);
      }

      rmul(AlphaW[i],NormV[i+1],D(k-1,i+1));
      for (j=0;j<=i;j++)
      {
	Complex whelp;
	mul(whelp,HV(j,i),D(k-1,j));
	add(AlphaW[i],AlphaW[i],whelp);
      }
      conj(AlphaW[i],AlphaW[i]);
    }
#if defined(CHECK_RHS)
/*     flaprintf(stdout,"Res AlphaV = %la"PPRE"vrte\nRes AlphaW = %la" */
/* 	  PPRE"vrte\n",1e-14,k,AlphaV,1,k,AV,1,1e-14,k,AlphaW,1,k,AW,1); */
    flaprintf(stdout,"Res AlphaV = %la"PPRE"vrneRes AlphaW = %la"
	  PPRE"vrne\n",1e-14,k,AlphaV,1,k,AV,1,1e-14,k,AlphaW,1,k,AW,1);
#endif
#if (defined(CHECK_RHS) || defined(DIRECT_RHS))
    pcopy (&k,AV,&inc1,AlphaV,&inc1);
    pcopy (&k,AW,&inc1,AlphaW,&inc1);
#endif

    PREC (ApplyDecomposition) (Work,Dinv,0,AlphaV);
    PREC (ApplyDecomposition) (Work,Dinv,1,AlphaW);

    CoefNorm = pasum(&k,AlphaV,&inc1);
    thelp = pasum(&k,AlphaW,&inc1);
    CoefNorm = max(CoefNorm,thelp);
  }

/* flaprintf(stdout,"AlphaV = %la"PPRE"vte\nAlphaW = %la"PPRE"vte\n",k,AlphaV,1,k,AlphaW,1); */
/* getchar(); */
  return CoefNorm;
#undef D
#undef HV
#undef HW
}

static Real PREC(GetBeta) (Complex *BetaV,Complex *BetaW,
      Complex *D__,finteger LdD,Real *NormV,Real *NormW,
      krylov *Kry,finteger k)
{
#define D(i,j) D__[(i)+(j)*LdD]
  if (KrylovSize(Kry) > 0)
  {
    Real thelp;
    Real CoefNorm;
    Complex Theta;

    /* Theta = D(0,k-1)/DOld */
    pdiv(Theta,D(0,k-1),LanczosLastD(Kry));

    /* H(n-1,n+k-1) = BetaV = NormW[0]*Theta */
    rmul(*BetaV,NormW[0],Theta);
    LanczosBetaMult(Kry)[LanczosRegularIndexRelativ(Kry,0)+k-1] = *BetaV;


    /* Theta = D(k-1,0)/DOld */
    pdiv(Theta,D(k-1,0),LanczosLastD(Kry));
    /* BetaW = conj(NormW[0]*Theta) */
    rmul(*BetaW,NormV[0],Theta);
    conj(*BetaW,*BetaW);

    CoefNorm = pabs(*BetaV)*LanczosNorm1BetaV(Kry);
    thelp = pabs(*BetaW)*LanczosNorm1BetaW(Kry);
    CoefNorm = max(CoefNorm,thelp);

    return CoefNorm;
  }
  else
    return 0.0;
#undef D
}

static void PREC(OrthogonalizeToCurrentBlock) (krylov *Kry,
      Complex *AlphaV,Complex *AlphaW,
      Complex *HV__,finteger LdHV,Complex *HW__,finteger LdHW,
      finteger k,finteger regular)
{
#define HV(i,j) HV__[(i)+(j)*LdHV]
#define HW(i,j) HW__[(i)+(j)*LdHW]

  if (regular)
  {
    finteger i;
    pcopy (&k,AlphaV,&inc1,&HV(0,k-1),&inc1);
    pcopy (&k,AlphaW,&inc1,&HW(0,k-1),&inc1);

    for (i=0;i<k;i++)
    {
      Complex whelp;
      neg(whelp,AlphaV[i]);
      paxpy (&KrylovNeq(Kry),&whelp,VN(i),&inc1,VN(k),&inc1);
    }

    for (i=0;i<k;i++)
    {
      Complex whelp;
      neg(whelp,AlphaW[i]);
      paxpy (&KrylovNeq(Kry),&whelp,WN(i),&inc1,WN(k),&inc1);
    }
  }
  else
  {
    /* Initialize corresponding entries in H. */
    pcopy (&k,&wzero,&inc0,&HV(0,k-1),&inc1);
    pcopy (&k,&wzero,&inc0,&HW(0,k-1),&inc1);

#ifdef ADDITIONAL_ORTHO
    {
      finteger i;
      for (i=0;i<k;i++)
      {
	Complex whelp;
	neg(whelp,AlphaV[i]);
	if ((pabs(whelp) > sqrt(plamch("P")))
	      && (pabs(whelp) < PREC(Exp4GetNormJacApproach) ()))
	{
/* 	  printf("               &&& Applying AlphaV[%d]=%e\n",i,AlphaV[i]); */
	  HV(i,k-1)=AlphaV[i];
	  paxpy (&KrylovNeq(Kry),&whelp,VN(i),&inc1,VN(k),&inc1);
	}
      }
      
      for (i=0;i<k;i++)
      {
	Complex whelp;
	neg(whelp,AlphaW[i]);
	if ((pabs(whelp) > sqrt(plamch("P")))
	      && (pabs(whelp) < PREC(Exp4GetNormJacApproach) ()))
	{
/* 	  printf("               &&& Applying AlphaW[%d]=%e\n",i,AlphaW[i]); */
	  HW(i,k-1)=AlphaW[i];
	  paxpy (&KrylovNeq(Kry),&whelp,WN(i),&inc1,WN(k),&inc1);
	}
      }
    }
#endif
#ifdef ADDITIONAL_ORTHO_II
    {
      finteger i;
      Real NAlphaV = pasum (&k,AlphaV,&inc1);
      Real NAlphaW = pasum (&k,AlphaW,&inc1);
      Real Mult = 1.0;

      if (2.0*max(NAlphaV,NAlphaW) > PREC(Exp4GetNormJacApproach) ())
	Mult = PREC(Exp4GetNormJacApproach) ()/(2.0*max(NAlphaV,NAlphaW));

      for (i=0;i<k;i++)
      {
	Complex whelp;
	rmul(whelp,Mult,AlphaV[i]);
	HV(i,k-1)=whelp;
	neg(whelp,whelp);
	paxpy (&KrylovNeq(Kry),&whelp,VN(i),&inc1,VN(k),&inc1);
      }
      
      for (i=0;i<k;i++)
      {
	Complex whelp;
	rmul(whelp,Mult,AlphaW[i]);
	HW(i,k-1)=whelp;
	neg(whelp,whelp);
	paxpy (&KrylovNeq(Kry),&whelp,WN(i),&inc1,WN(k),&inc1);
      }
    }
#endif
#ifdef ARNOLDI_ORTHO
    {
      Complex whelp;
      int i;

      for (i=0;i<k;i++)
      {
	pdot (&whelp,&KrylovNeq(Kry),VN(i),&inc1,VN(k),&inc1);
	
	HV(i,k-1) = whelp;
	neg(whelp,whelp);
	paxpy (&KrylovNeq(Kry),&whelp,VN(i),&inc1,VN(k),&inc1);
      }

      for (i=0;i<k;i++)
      {
	pdot (&whelp,&KrylovNeq(Kry),WN(i),&inc1,WN(k),&inc1);
	
	HW(i,k-1)=whelp;
	neg(whelp,whelp);
	paxpy (&KrylovNeq(Kry),&whelp,WN(i),&inc1,WN(k),&inc1);
      }
    }
#endif
#ifdef ARNOLDI_ORTHO_II
    {
      Complex whelp;
      int i;

      for (i=max(0,k-ARNOLDI_BAND);i<k;i++)
      {
	pdot (&whelp,&KrylovNeq(Kry),VN(i),&inc1,VN(k),&inc1);
	
	HV(i,k-1) = whelp;
	neg(whelp,whelp);
	paxpy (&KrylovNeq(Kry),&whelp,VN(i),&inc1,VN(k),&inc1);
      }

      for (i=max(0,k-ARNOLDI_BAND);i<k;i++)
      {
	pdot (&whelp,&KrylovNeq(Kry),WN(i),&inc1,WN(k),&inc1);
	
	HW(i,k-1)=whelp;
	neg(whelp,whelp);
	paxpy (&KrylovNeq(Kry),&whelp,WN(i),&inc1,WN(k),&inc1);
      }
    }
#endif
  }
#undef HV
#undef HW
}

static void PREC(OrthogonalizeToLastBlock) (krylov *Kry,Complex BetaV,
      Complex BetaW,finteger k)
{
  if (KrylovSize(Kry) > 0)
  {
    Complex whelp;
    finteger ihelp = KrylovSize(Kry)-1;
    finteger lastsize = LanczosRegularIndexRelativ(Kry,0)
      -LanczosRegularIndexRelativ(Kry,-1);

    /* initialize the ihelp't column with zeros */
    pcopy (&LanczosRegularIndexRelativ(Kry,-1),&wzero,&inc0,
	  &KrylovHent(Kry,0,ihelp+k),&inc1);

    if (lastsize == 1)
    {
      /* H(n-1,n+k-1) = BetaV */
      KrylovHent(Kry,ihelp,ihelp+k) = BetaV;
    }
    else
    {
      pcopy (&lastsize,LanczosBetaTemp(Kry)+LanczosRegularIndexRelativ(Kry,-1),
	    &inc1,
	    &KrylovHent(Kry,LanczosRegularIndexRelativ(Kry,-1),ihelp+k),&inc1);
      pscal (&lastsize,&BetaV,
	    &KrylovHent(Kry,LanczosRegularIndexRelativ(Kry,-1),ihelp+k),&inc1);
    }
    /* VN(k) -= BetaV*VN(k) */
    neg(whelp,BetaV);
    paxpy (&KrylovNeq(Kry),&whelp,VN(-1),&inc1,VN(k),&inc1);

    /* WN(k) -= BetaW*WN(k) */
    neg(whelp,BetaW);
    paxpy (&KrylovNeq(Kry),&whelp,WN(-1),&inc1,WN(k),&inc1);
  }
}

static void PREC(CalculateTemporary) (WorkSpace Work,krylov *Kry,
      Decomposition *Dinv,finteger k)
{
  int i;
  Complex *Beta = LanczosBetaTemp(Kry)+LanczosRegularIndexRelativ(Kry,0);

  /* calculate help W vector */
  for (i=0;i<k-1;ComplexSetReal(Beta[i],0.0),i++);
  ComplexSetReal(Beta[k-1],1.0);
  PREC (ApplyDecomposition) (Work,Dinv,1,Beta);
  pscal (&KrylovNeq(Kry),Beta+k-1,WN(k-1),&inc1);
  for (i=0;i<k-1;i++)
    paxpy (&KrylovNeq(Kry),Beta+i,WN(i),&inc1,WN(k-1),&inc1);

  LanczosNorm1BetaW(Kry) = pasum (&k,Beta,&inc1);

  /* calculate help V vector */
  for (i=0;i<k-1;ComplexSetReal(Beta[i],0.0),i++);
  ComplexSetReal(Beta[k-1],1.0);
  PREC (ApplyDecomposition) (Work,Dinv,0,Beta);
  pscal (&KrylovNeq(Kry),Beta+k-1,VN(k-1),&inc1);
  for (i=0;i<k-1;i++)
    paxpy (&KrylovNeq(Kry),Beta+i,VN(i),&inc1,VN(k-1),&inc1);

  LanczosNorm1BetaV(Kry) = pasum (&k,Beta,&inc1);

/* printf("Norm1BetaV = %e, NormTempV = %e\n",LanczosNorm1BetaV(Kry), */
/*       pnrm2(&KrylovNeq(Kry),VN(k-1),&inc1)); */
/* printf("Norm1BetaW = %e, NormTempW = %e\n",LanczosNorm1BetaW(Kry), */
/*       pnrm2(&KrylovNeq(Kry),WN(k-1),&inc1)); */
}


static void PREC(UpdateD) (krylov *Kry,Complex *D__,finteger LdD,
      Real *NormV,Real *NormW,
      Complex *HV__,finteger LdHV,Complex *HW__,finteger LdHW,
      finteger k,int V0_EQ_W0)
{
#define D(i,j) D__[(i)+(j)*LdD]
#define HV(i,j) HV__[(i)+(j)*LdHV]
#define HW(i,j) HW__[(i)+(j)*LdHW]

  if ((KrylovSize(Kry) == 0) && V0_EQ_W0)
  {
    /* if we'r in the first step w == v and || v || = 1 */
    ComplexSetReal(D(0,0),1.0);
  }
  else
  {
    int i;
    
#if (defined(CHECK_D) || defined(DIRECT_D))
    Complex D[(k+1)*(k+1)];
    {
      int i,j;
      for(i=0;i<=k;i++)
	for (j=0;j<k;j++)
	  pdot (&D[i+(k+1)*j],&KrylovNeq(Kry),WN(i),&inc1,VN(j),&inc1);
    }
#endif

    /* calculate new diagonal value, it is needed for calculating beta coefs */
    pdot(&D(k-1,k-1),&KrylovNeq(Kry),WN(k-1),&inc1,VN(k-1),&inc1);

    /* get the values of the sub- and super-diagonal from w'*A*v calculated by
     * UpdateD_II */
    if (k > 1)
    {
      int j;

      for (j=0;j<=k-2;j++)
      {
	Complex whelp;
	mul(whelp,HV(j,k-2),D(k-2,j));
	sub(D(k-2,k-1),D(k-2,k-1),whelp);
      }
      D(k-1,k-2) = D(k-2,k-1);
      rdiv(D(k-2,k-1),D(k-2,k-1),NormV[k-1]);


      for (j=0;j<=k-2;j++)
      {
	Complex whelp;
	mul(whelp,HV(j,k-2),D(k-2,j));
	add(D(k-1,k-2),D(k-1,k-2),whelp);
      }

      for (j=0;j<=k-2;j++)
      {
	Complex whelp;
	conj(whelp,HW(j,k-2));
	mul(whelp,whelp,D(j,k-2));
	sub(D(k-1,k-2),D(k-1,k-2),whelp);
      }
      rdiv(D(k-1,k-2),D(k-1,k-2),NormW[k-1]);
    }
    
    /* calculate the last line of D by using the Hankel structure of the 
     * unscaled algorithem (that means rescaling of the coefficients
     * has to be done) */
    for (i=0;i<k-2;i++)
    {
      int j;
      rmul(D(k-1,i),NormV[i+1],D(k-2,i+1));

      for (j=0;j<=i;j++)
      {
	Complex whelp;
	mul(whelp,HV(j,i),D(k-2,j));
	add(D(k-1,i),D(k-1,i),whelp);
      }

      for (j=0;j<=k-2;j++)
      {
	Complex whelp;
	conj(whelp,HW(j,k-2));
	mul(whelp,whelp,D(j,i));
	sub(D(k-1,i),D(k-1,i),whelp);
      }

      rdiv(D(k-1,i),D(k-1,i),NormW[k-1]);
    }

    /* calculate the last column of D by using the Hankel structure of the 
     * unscaled algorithem. */
    for (i=0;i<k-2;i++)
    {
      int j;
      rmul(D(i,k-1),NormW[i+1],D(i+1,k-2));

      for (j=0;j<=k-2;j++)
      {
	Complex whelp;
	mul(whelp,HV(j,k-2),D(i,j));
	sub(D(i,k-1),D(i,k-1),whelp);
      }

      for (j=0;j<=i;j++)
      {
	Complex whelp;
	conj(whelp,HW(j,i));
	mul(whelp,whelp,D(j,k-2));
	add(D(i,k-1),D(i,k-1),whelp);
      }

      rdiv(D(i,k-1),D(i,k-1),NormV[k-1]);
    }
#if defined(CHECK_D)
    flaprintf(stdout,"Err D = \n%la" PPRE "gerne\n",1e-17,k,k,D__,LdD,
	  k,k,D,k+1);
#endif
#if defined(DIRECT_D)
    {
      finteger ihelp = k+1;
      placpy ("A",&k,&k,D,&ihelp,D__,&LdD);
    }
#endif
/*     flaprintf(stdout,"D = \n%la"PPRE"gee\n",k,k,D__,LdD); */
/* getchar(); */
  }
#undef D
#undef HV
#undef HW
}

static int PREC(Normalize) (Real *NormV,Real *NormW,
      krylov *Kry,finteger k)
{
  finteger ihelp = KrylovSize(Kry) + k - 1;

  if (ihelp >= 0)
  {
    int FoundInvariantVSubspace = 0;
    /* get norm of V(KrylovSize(Kry)+k) */
    NormV[k] = pnrm2 (&KrylovNeq(Kry),VN(k),&inc1);
    /* get norm of W(KrylovSize(Kry)+k) */
    NormW[k] = pnrm2 (&KrylovNeq(Kry),WN(k),&inc1);

    /* check for invariant subspaces */
    if (NormV[k] < plamch("eps")) FoundInvariantVSubspace = 1;
    /* check for invariant subspaces, if W is invariant but V not return 1
     * to force a restart with a random vector */
    if ((NormW[k] == 0.0) && !FoundInvariantVSubspace) return 1;

    /* normalize V(KrylovSize(Kry)+k) */
    prscl (&KrylovNeq(Kry),NormV+k,VN(k),&inc1);
    /* normalize W(KrylovSize(Kry)+k) */
    prscl (&KrylovNeq(Kry),NormW+k,WN(k),&inc1);
    
    /* set coresponding entry in matrix H */ 
    /* write the KrylovSize(Kry)+k - th line of H */
    if ((KrylovSize(Kry)+k) < KrylovLdSmall(Kry))
    {
      ComplexSetReal(KrylovHent(Kry,KrylovSize(Kry)+k,ihelp),NormV[k]);
      pcopy (&ihelp,&wzero,&inc0,&KrylovHent(Kry,KrylovSize(Kry)+k,0),
	    &KrylovLdSmall(Kry));
    }
    
    /* if the recursion is 'very regular' the subdiagonal of H is
     * a lower bound for || A || */
    {
      finteger LastReg = LanczosRegularIndexRelativ(Kry,-1);
      finteger ihelp = KrylovSize(Kry)+k-LastReg;
      if (pasum(&ihelp,&KrylovHent(Kry,LastReg,KrylovSize(Kry)+k-1),&inc1)
	    < NormV[k])
	PREC(Exp4SetNormJacApproach) (NormV[k]);
    }
    
#ifdef STORE_FULL
    PREC(StoreLanczosVect) (Kry,VN(k),WN(k),KrylovSize(Kry)+k);
#endif
  }
  return 0;
}

static Real PREC(DecompositeMatrix) (WorkSpace Work,
      Decomposition *Minv,finteger k,Complex *m,finteger ldm)
{
  if (k > 1) {
#ifdef USE_SVD
#if (defined(DOUBLE_PRECISION) || defined(SINGLE_PRECISION))
    Complex *work;
    finteger info;
    finteger lwork = GetRemainingWorkSpaceSize(Work)/sizeof(Complex);
    lwork = max(lwork,sizeof(Complex)*5*k);
    DistributeWorkspace(work,lwork,Work);
    Minv->n = k;

    placpy ("A",&k,&k,m,&ldm,Minv->U,&Minv->n);
    
    pgesvd ("O","A",&k,&k,Minv->U,&k,Minv->S,NULL,&inc1,Minv->VT,&k,
	  work,&lwork,&info);
    /* if a serious error occurs */
    lapack_err(info,"DecompositeMatrix","gesvd");
#else
    Complex *work;
    Real *rwork;
    finteger info;
    finteger lwork;
    DistributeWorkspace(rwork,5*k*sizeof(Real),Work);
    lwork = GetRemainingWorkSpaceSize(Work)/sizeof(Complex);
    lwork = max(lwork,sizeof(Complex)*3*k);
    DistributeWorkspace(work,lwork,Work);
    Minv->n = k;

    placpy ("A",&k,&k,m,&ldm,Minv->U,&Minv->n);
    
    pgesvd ("O","A",&k,&k,Minv->U,&k,Minv->S,NULL,&inc1,Minv->VT,&k,
	  work,&lwork,rwork,&info);
    /* if a serious error occurs */
    lapack_err(info,"DecompositeMatrix","gesvd");
#endif
    return (Minv->S)[k-1];
#else
    Real DiagMin;
    int i;
    finteger info;
    Minv->n = k;

    placpy ("A",&k,&k,m,&ldm,Minv->Data,&Minv->n);
#ifdef USE_QR_DECOMPOSITION
    {
      Complex *work;
      DistributeWorkspace(work,sizeof(Complex)*3*k,Work);
      for (i=0;i<k;Minv->Pivot[i++]=0);
      pgeqpf(&k,&k,Minv->Data,&Minv->n,Minv->Pivot,Minv->Tau,work,&info);

      /* if a serious error occurs */
      lapack_err(info,"DecompositeMatrix","geqpf");
    }
#else
    /* copy and get LU decomposition */
    pgetrf (&Minv->n,&Minv->n,Minv->Data,&Minv->n,Minv->Pivot,&info);

    /* if matrix is exact singular */
    if (info > 0) return 0.0;

    /* if a serious error occurs */
    lapack_err(info,"DecompositeMatrix","getrf");
#endif

    /* get smallest diagonal value */
    DiagMin = pabs(*(Minv->Data));
    for (i=1;i<k;i++) {
      Real thelp = pabs(*(Minv->Data+(k+1)*i));
      if (thelp < DiagMin)
	DiagMin = thelp;
    }

    return DiagMin;
#endif
  } else {
#ifdef USE_SVD
    (Minv->U)[0] = *m;
#else
    (Minv->Data)[0] = *m;
#endif
    return pabs(*m);
  }
}

static finteger PREC (ApplyDecomposition) (WorkSpace Work,
      Decomposition *Minv,int CmplxTransp,Complex *b)
{
#ifdef USE_SVD
  if (Minv->n > 1) {
    if (CmplxTransp) {
      int i;
      Complex *work;
      DistributeWorkspace(work,sizeof(Complex)*Minv->n,Work);

      pgemv ("N",&Minv->n,&Minv->n,&wone,Minv->VT,&Minv->n,b,&inc1,
	    &wzero,work,&inc1);
      
      for (i=0; i<Minv->n; i++) {
	rdiv(work[i],work[i],(Minv->S)[i]);
      }

      pgemv ("N",&Minv->n,&Minv->n,&wone,Minv->U,&Minv->n,work,&inc1,
	    &wzero,b,&inc1);
    } else {
      int i;
      Complex *work;
      DistributeWorkspace(work,sizeof(Complex)*Minv->n,Work);
      
      pgemv ("C",&Minv->n,&Minv->n,&wone,Minv->U,&Minv->n,b,&inc1,
	    &wzero,work,&inc1);
      
      for (i=0; i<Minv->n; i++) {
	rdiv(work[i],work[i],(Minv->S)[i]);
      }
      
      pgemv ("C",&Minv->n,&Minv->n,&wone,Minv->VT,&Minv->n,work,&inc1,
	    &wzero,b,&inc1);
    }
    return ((Minv->S)[Minv->n - 1] != 0.0);
  } else {
    pdiv(b[0],b[0],(Minv->U)[0]);
    return (!ZERO((Minv->U)[0]));
  }
#else
  finteger info;
#ifdef USE_QR_DECOMPOSITION
  if (CmplxTransp)
  {  
    flogical flag = FTRUE;
    Complex *work;
    DistributeWorkspace(work,sizeof(Complex)*Minv->n,Work);

    plapmt (&flag,&inc1,&Minv->n,b,&inc1,Minv->Pivot);

    ptrsv ("U","C","N",&Minv->n,Minv->Data,&Minv->n,b,&inc1);
  
    porm2r ("L","N",&Minv->n,&inc1,&Minv->n,Minv->Data,&Minv->n,Minv->Tau,
	  b,&Minv->n,work,&info);
    lapack_err(info,"ApplyDecomposition","orm2r");
  }
  else
  {
    flogical flag = FFALSE;
    Complex *work;
    DistributeWorkspace(work,sizeof(Complex)*Minv->n,Work);

#if (defined(SINGEL_PRECISION) || defined(DOUBLE_PRECISION))
    porm2r ("L","T",&Minv->n,&inc1,&Minv->n,Minv->Data,&Minv->n,Minv->Tau,
	  b,&Minv->n,work,&info);
#else
    porm2r ("L","C",&Minv->n,&inc1,&Minv->n,Minv->Data,&Minv->n,Minv->Tau,
	  b,&Minv->n,work,&info);
#endif
    lapack_err(info,"ApplyDecomposition","orm2r");
     
    ptrsv ("U","N","N",&Minv->n,Minv->Data,&Minv->n,b,&inc1);
    
    plapmt (&flag,&inc1,&Minv->n,b,&inc1,Minv->Pivot);
  }
#else
  if (CmplxTransp)
    pgetrs ("C",&Minv->n,&inc1,Minv->Data,&Minv->n,Minv->Pivot,b,&Minv->n,
	  &info);
  else
    pgetrs ("N",&Minv->n,&inc1,Minv->Data,&Minv->n,Minv->Pivot,b,&Minv->n,
	  &info);
#endif
  return info;
#endif
}

static finteger PREC(BestBlockSize) (Real *CoefNorm,finteger k)
{
  finteger i;
  finteger Best = 0;
  Real BestVal = plamch("O");

  for (i=1;i<=k;i++)
  {
    /* negative CoefNorm values indicate very singular D */
    if ((0.0 <= CoefNorm[i]) && (CoefNorm[i] < BestVal))
      Best = i, BestVal = CoefNorm[i];
  }
  
  /* if only very singular blocks are buildt return zero */
  if (Best == 0) return 0;

  /* all Blocks with CoefNorm between BestVal and COEF_NORM_TOLERANCE*BestVal
   * are good enough, so take the largest one to minimize execution time. */
/*   BestVal *= max(1.0,COEF_NORM_TOLERANCE); */

/*   for (i=k;i>0;i--) */
/*   { */
    /* negative CoefNorm values indicate very singular D */
/*     if ((0.0 <= CoefNorm[i]) && (CoefNorm[i] <= BestVal)) */
/*     { */
/*       Best = i; */
/*       break; */
/*     } */
/*   } */

  return Best;
}

/* undefining the VN and WN defines for building the Krylov space */
#undef VN
#undef WN
#undef InvalidateVN


/**********************************************************************/
/*                                                                    */
/*                rebuilding the krylov base                          */
/*                                                                    */
/**********************************************************************/

/* the last V vectors which are here from building the krylov space */
#define STOREDVN(I)  (LanczosV(Kry,(I)%LanczosVStride(Kry)))

/* the rebuilding process will be done in the memory for the W vectors */
#define REBUILDVN(I) (LanczosW(Kry,(I)%LanczosWStride(Kry)))

/* vectors for rebuilding the temporary vectors in the stored V can be
 * in the V or W workspace, this macro looks for the right one */
#define EARLYVN(Last,I) (((I)<(Last)) ? REBUILDVN(I):STOREDVN(I))



static void PREC(Exp4KrylovCombineLinearLanczos) (Complex *x,
      finteger ldx,Complex *Start,krylov *Kry,
      Complex *zz,finteger ldz,finteger fac_t,int HaveOldV,int dummy)
     /* This routine rebuilds the V vectors from the stored recursion
      * coeficients.
      * If HaveOldV is set to 1 the vectors stored in V are the last V
      * vectors of the Krylov base, so they can be reused. If HaveOldV
      * is set to 0 the whole base must be rebuilt.
      */
{
#define z(I,J)      zz[(I) + ldz*(J)]
#define X(I,J)      x[(I) + ldx*(J)]

#ifdef STORE_FULL
/*   PREC(PrintVW) (Kry); */
/* PREC(BiOrth) (Kry,0,1e-10); */
/* PREC(testAV_VH) (Kry,-1,1); */
/* getchar(); */
#endif

#if (DEBUG > LOCAL_DEBUG_OFFSET)
  {
    int i;
    fprintf(DEBUG_FILE,"       Approach to || A || = %e\n",
	  PREC(Exp4GetNormJacApproach) ());
    fprintf(DEBUG_FILE,"       Building Linear Combination, size = %d, "
	  "Blocks between: ",KrylovSize(Kry));
    for (i=1;i<=KrylovPrivat(Kry).Lanczos.RegularIndexIndex;i++)
      if (KrylovPrivat(Kry).Lanczos.RegularIndex[i] !=
	    KrylovPrivat(Kry).Lanczos.RegularIndex[i-1]+1)
	printf("(%d %d), ",KrylovPrivat(Kry).Lanczos.RegularIndex[i-1],
	      KrylovPrivat(Kry).Lanczos.RegularIndex[i]);
    putchar('\n');
  }
#endif

  /* initialize the array if needed */
  {
    finteger ihelp = fac_t * ldx;
    pcopy (&ihelp,&wzero,&inc0,&X(0,0),&inc1);
  }
  
  /* take the start vector for the linear combination */
  {
    Complex rNormv;
    ComplexSetReal(rNormv,1.0/KrylovNormv(Kry));
    pger(&KrylovNeq(Kry),&fac_t,&rNormv,Start,&inc1,&z(0,0),&ldz,&X(0,0),&ldx);
    Exp4Message8((DEBUG_FILE,"Taking startvector for Linearcombination\n"));
  }

  if (KrylovSize(Kry) > 1) {
    /* helper variable */
    Complex whelp;
    finteger i;
    /* the last V needed for linear combination */
    finteger LastVToCalculate =  KrylovSize(Kry);
    /* the actual block number to work with */
    finteger BlockNumber = 0;
    /* the size of the last block */
    finteger LastBlockSize = 0;
    /* the size of the actual block */
    finteger ThisBlockSize;
    /* the actual V index to recompute */
    finteger CurrentVIndex = 1;
    /* the actual regular Index */
    finteger RegInd = 0;
    /* the Index of the actual temporary vector to use */
    finteger TempIndex = -1;


    /* if the V workspace holds the last V vectors calculated while building
     * the Lanczos base, reuse them (that means only rebuild the first ones) */
    if (HaveOldV)
      LastVToCalculate = min(LanczosLastValidV(Kry),LastVToCalculate);
    Exp4Message8((DEBUG_FILE,"LastVToCalculate = %d\n", LastVToCalculate));

    /* copy and rescale the first vector */
    pcopy (&KrylovNeq(Kry),Start,&inc1,REBUILDVN(0),&inc1);
    prscl (&KrylovNeq(Kry),&KrylovNormv(Kry),REBUILDVN(0),&inc1);
#ifdef STORE_FULL
    PREC(StoreLanczosVect) (Kry,NULL,REBUILDVN(0),0);
#endif
    ThisBlockSize = LanczosReadRegularIndex(Kry,1)
      - LanczosReadRegularIndex(Kry,0);

    /* rebuild all vectors up to LastVToCalculate, the last vectors are in
     * the V workspace */
    while (CurrentVIndex < LastVToCalculate)
    {
      /* the actual column index of H to process */
      finteger HColumn = CurrentVIndex - 1;
      /* the Position within the block */
      finteger BlockPosition = CurrentVIndex - RegInd;

      /* multiply with the Jacobian */
      PREC(Exp4Mv) (KrylovNeq(Kry),REBUILDVN(CurrentVIndex),
	    REBUILDVN(CurrentVIndex-1),1);
      
      /* apply the beta coeficients */
      if (1 == LastBlockSize)
      {
	/* Code for regular last blocks */
	neg(whelp,KrylovHent(Kry,TempIndex,HColumn));
	paxpy (&KrylovNeq(Kry),&whelp,REBUILDVN(TempIndex),&inc1,
	      REBUILDVN(CurrentVIndex),&inc1);
      }
      else if (1 < LastBlockSize)
      {
	/* Code for nonregular last blocks */
	neg(whelp,LanczosBetaMult(Kry)[HColumn]);
	paxpy (&KrylovNeq(Kry),&whelp,REBUILDVN(TempIndex),&inc1,
	      REBUILDVN(CurrentVIndex),&inc1);
      }
      
      /* apply the alpha coeficients */
      for (i=0;i<BlockPosition;i++)
      {
	finteger Ind = RegInd+i;
	neg(whelp,KrylovHent(Kry,Ind,HColumn));
	if (!ComplexIsZero(whelp))
	  paxpy (&KrylovNeq(Kry),&whelp,REBUILDVN(Ind),&inc1,
		REBUILDVN(CurrentVIndex),&inc1);
      }

      /* normalize the current V (KrylovHent(Kry,k+1,k) is real anyway) */
      prscl (&KrylovNeq(Kry),&Re(KrylovHent(Kry,CurrentVIndex,HColumn)),
	    REBUILDVN(CurrentVIndex),&inc1);
      
#ifdef STORE_FULL
      PREC(StoreLanczosVect) (Kry,NULL,REBUILDVN(CurrentVIndex),CurrentVIndex);
#endif
      /* add the rebuilt vector to the result */
      pger (&KrylovNeq(Kry),&fac_t,&wone,REBUILDVN(CurrentVIndex),&inc1,
	    &z(CurrentVIndex,0),&ldz,&X(0,0),&ldx);
      Exp4Message8((DEBUG_FILE,"Taking rebuilt vector V(%d) for "
	    "Linearcombination\n",CurrentVIndex));

      /* close the actual block */
      if ((RegInd+ThisBlockSize) <= CurrentVIndex)
      {
	/* update the index of the temporary vector */
	TempIndex += ThisBlockSize;

	Exp4Message9((DEBUG_FILE,"Closing Block of size %d at k = %d\n",
		ThisBlockSize, CurrentVIndex));

	/* calculate the Temporary only if it is really used to
	 * calculate a new V vector with it */
	if ((1 < ThisBlockSize) && (CurrentVIndex <= KrylovSize(Kry)))
	{
	  Exp4Message9((DEBUG_FILE,"Rebuilding temporary vector V(%d)\n",
		TempIndex));

	  pscal (&KrylovNeq(Kry),&LanczosBetaTemp(Kry)[TempIndex],
		REBUILDVN(TempIndex),&inc1);
	  for (i=0;i<(ThisBlockSize-1);i++)
	    paxpy (&KrylovNeq(Kry),&LanczosBetaTemp(Kry)[RegInd+i],
		  REBUILDVN(RegInd+i),&inc1,REBUILDVN(TempIndex),&inc1);
	}

	/* update indices for the block structure */
	BlockNumber++;
	RegInd = LanczosReadRegularIndex(Kry,BlockNumber);
	LastBlockSize = ThisBlockSize;
	ThisBlockSize = LanczosReadRegularIndex(Kry,BlockNumber+1)-RegInd;
      }

      CurrentVIndex++;
    }

    /* now recalculate these vectors in the V array which are overwritten
     * by a temporary vector, this is done from last one down to the first
     * one so the temporary vectors in V are reused */
    for (i=KrylovPrivat(Kry).Lanczos.RegularIndexIndex;i>BlockNumber;i--)
    {
      /* calculate the size the the current block */
      ThisBlockSize = LanczosReadRegularIndex(Kry,i) -
	LanczosReadRegularIndex(Kry,i-1);

      /* if this block is a real block (size > 1) recalculate the
       * vector overwritten by the temporary vector */
      if (ThisBlockSize > 1)
      {
	finteger HColumn;
	finteger j;
	/* the index of the temporary vector to rebuild */
	TempIndex = LanczosReadRegularIndex(Kry,i) - 1;
	/* the column where to find the corresponding coeficients */
	HColumn = TempIndex - 1;
	/* the size of the last block */
	LastBlockSize = LanczosReadRegularIndex(Kry,i-1)-
	  LanczosReadRegularIndex(Kry,i-2);
	
	/* stop recalculating if the vector to bebuild is never needed */
	if (TempIndex < LastVToCalculate) break;

	Exp4Message8((DEBUG_FILE,"Recalculating V(%d) previously overwritten "
	      "by temporary vector\n",TempIndex));

	Exp4Message9((DEBUG_FILE,"Taking %s vector V(%d) for multiplikation "
	      "with A\n",
	      (EARLYVN(LastVToCalculate,TempIndex-1) == REBUILDVN(TempIndex-1))
	      ?"rebuilt":"cached",TempIndex-1));

	/* multiply with the Jacobian */
	PREC(Exp4Mv) (KrylovNeq(Kry),STOREDVN(TempIndex),
	      EARLYVN(LastVToCalculate,TempIndex-1),1);
      

	/* apply the beta coeficients */
	if (1 == LastBlockSize)	{
	  /* Code for regular last blocks */
	  finteger Ind = TempIndex-ThisBlockSize;
	  neg(whelp,KrylovHent(Kry,Ind,HColumn));
	  paxpy (&KrylovNeq(Kry),&whelp,EARLYVN(LastVToCalculate,Ind),&inc1,
		STOREDVN(TempIndex),&inc1);
	} else if (1 < LastBlockSize) {
	  /* Code for nonregular last blocks */
	  finteger Ind = TempIndex-ThisBlockSize;

	  /* If the last rebuilt vector is block-end Vector the temporary
	     vector which is expected in this place will now be rebuilt.
	     This calculation can not be done above, because the original
	     vector at this location may be needed for multiplication
	     with A. */
/* $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$4 */
	  if (Ind+1 == CurrentVIndex) {
	    int tmpidx;
	    Exp4Message9((DEBUG_FILE,"Rebuilding temporary vector V(%d) "
		  "needed for building temporary vector V(%d)\n",
		  Ind,TempIndex));

	    pscal (&KrylovNeq(Kry),&LanczosBetaTemp(Kry)[Ind],
		  REBUILDVN(Ind),&inc1);
	    for (tmpidx=LastBlockSize-1;0<tmpidx;tmpidx--)
	      paxpy(&KrylovNeq(Kry),&LanczosBetaTemp(Kry)[Ind-tmpidx],
		    REBUILDVN(Ind-tmpidx),&inc1,REBUILDVN(Ind),&inc1);
	  }

	  neg(whelp,LanczosBetaMult(Kry)[HColumn]);
	  paxpy (&KrylovNeq(Kry),&whelp,EARLYVN(LastVToCalculate,Ind),&inc1,
		STOREDVN(TempIndex),&inc1);

	  Exp4Message9((DEBUG_FILE,"Taking %s temporary vector V(%d) for "
		"calculating new temporary vector V(%d)\n",
		(EARLYVN(LastVToCalculate,Ind) == REBUILDVN(Ind))
		?"rebuilt":"cached",Ind,TempIndex));
	}

	/* apply the alpha coeficients */
	for (j=0;j<ThisBlockSize-1;j++)
	{
	  finteger Ind = TempIndex-ThisBlockSize+1+j;
	  neg(whelp,KrylovHent(Kry,Ind,HColumn));

	  if (!ComplexIsZero(whelp))
	    paxpy (&KrylovNeq(Kry),&whelp,EARLYVN(LastVToCalculate,Ind),&inc1,
		  STOREDVN(TempIndex),&inc1);
	}

	/* normalize the current V (KrylovHent(Kry,k+1,k) is real anyway) */
	prscl (&KrylovNeq(Kry),&Re(KrylovHent(Kry,TempIndex,HColumn)),
	      STOREDVN(TempIndex),&inc1);
      }
    }

    /* use the stored (and partially rebuilt) vectors for the
     * linear kombination */
    while (CurrentVIndex < KrylovSize(Kry))
    {
#ifdef STORE_FULL
      PREC(StoreLanczosVect) (Kry,NULL,STOREDVN(CurrentVIndex),CurrentVIndex);
#endif
      /* add the rebuilt vector to the result */
      pger (&KrylovNeq(Kry),&fac_t,&wone,STOREDVN(CurrentVIndex),&inc1,
	    &z(CurrentVIndex,0),&ldz,&X(0,0),&ldx);
      Exp4Message8((DEBUG_FILE,"Taking cached vector V(%d) for "
	    "Linearcombination\n",CurrentVIndex));
      CurrentVIndex++;
    }
  }

#ifdef STORE_FULL
  PREC(CompareVW) (Kry);
#endif

#undef z
#undef X
}

#undef VN
