/*************************************************************************/
/* Copyright (C) 1995-1997 Marlis Hochbruck, Christian Lubich, Hubert    */
/* Selhofer. All rights reserved.                                        */
/*                                                                       */
/* C code written by Mathias Froehlich.                                  */
/*                                                                       */
/* This code is part of a copyrighted package. For details, see the file */
/* "COPYING" in the top-level directory.                                 */
/*************************************************************************/

/*
 * This code is part of EXP4.
 *
 *  This functions computes the Krylov approximation to
 *     phi(t*A)v with phi(z) = (exp(z)-1)/z.
 * The matrix A is given implicitly.
 * The convergence criterion is based on a generalized residual.
 */

#include <stdlib.h>
#include <stdio.h>
#include <float.h>
#include <math.h>
#include "ftypes.h"
#include "fblas.h"
#include "exp4.h"
#include "prec.h"
#include "errors.h"

/* #define LOCAL_DEBUG_LEVEL 20 */
#define LOCAL_DEBUG_OFFSET 5
#include "debug.h"

#include "krylov.h"
#include "scnorm.h"
#include "hbounds.h"

#include "phikry.h"

#ifdef max
#undef max
#endif
#define max(a,b) (((a)<(b))?(b):(a))


/* if EXISTS_ISNAN is defined in the makefile take the libm routine isnan
 * to determine if a value is Not a Number. if this is not defined, try to
 * find out by comparing the value with itself. (on some machines NaN != NaN
 * is true) If nothing does work, exp4 might hang in such a case. */
#if (!defined(EXISTS_ISNAN) &&  !defined(isnan))
#define isnan(val) ((val) != (val))
#endif

#define TOUGH_BOUND        1e20
#define TOUGH_RED_FAC      1.0/2.0

static finteger inc0 = 0;
static finteger inc1 = 1;

static Complex wzero = ComplexConst(0.0);

static int DisableCaching = 0;

void PREC(Exp4SetDisableCaching) (Exp4Boolean Val)
{
  if (Val == Exp4True)
    DisableCaching = 1;
  else
    DisableCaching = 0;
}

size_t PREC(Exp4SizePhiKryWork) (krylov *Kry,finteger fac_t)
{
  size_t SharedSize,ResidentSize,SharedSizeSec,SelfSize;
  /* Size for Z */
  SelfSize = GetWorkSpaceSize(sizeof(Complex)*fac_t*KrylovLdSmall(Kry));
  /* Size for the KrylovSpace routines */
  KrylovSizeWork (Kry,&SharedSize,&ResidentSize);
  if (KrylovHasSecondary(Kry)) {
    KrylovActiveSecondary(Kry);
    KrylovSizeWork (Kry,&SharedSizeSec,&ResidentSize);
    KrylovActivePrimary(Kry);
    SharedSize = max(SharedSize,SharedSizeSec);
  }
  /* Size for calculating phim, stored in the Krylov handle for Lanczos and
   * Cgs, both use all possible Worspace for caching of V vectors. */
  KrylovReserveSize(Kry) = PhiZSizeWork(Kry,KrylovMaxSize(Kry));

  SelfSize += max(SharedSize,KrylovReserveSize(Kry));
  KrylovReserveSize(Kry) = SelfSize;
  return SelfSize+ResidentSize;
}

static Real PREC(GetResidual) (finteger KrylovSize,Real h,
      Real gamma,finteger fac_t,Real sc_norm,
      Real errw,Complex *Z,finteger LdZ)
{
#define z(I,J)      Z[(I) + LdZ*(J)]
  int i;
  Real Res = pabs(z(KrylovSize-1,fac_t-1))*((Real)fac_t);

  for (i=0;i<fac_t-1;i++)
  {
    Real Tmp = pabs(z(KrylovSize-1,i))*((Real)(i+1));
    if (Res < Tmp) Res = Tmp;
  }

  Res *= h*sc_norm*h*h*gamma*max(1.0,errw);

  if (KrylovSize == 1) Res *= 5.0;

  Exp4Message2((DEBUG_FILE,"KrylovStep: %3d, || V(%3d) || = %9.2e,"
	" res= %9.2e\n",(int)(KrylovSize),(int)(KrylovSize+1),
	(double)sc_norm,(double)Res));

  return Res;
#undef z
}

static void PREC(PhiKrylovCombineExistingBase) (WorkSpace Work,
      Complex *X,finteger LdX,krylov *Kry,Complex *Start,
      Real h,Real gamma,finteger fac_t,finteger neq)
/* PhiKrylovCombineExisting calculates an approximation to the matrix
 * function phi(h*A)*v.
 *
 * parameters:
 *
 * Work    A pointer to the workspace needed by PhiKrylov
 *
 * x       A pointer to the result matrix including fac_t vectors of the
 *
 * Kry
 *
 * Start   the vector to multiply with the matrix function
 *
 * h       the multiplication factor for the matrix A.
 *
 * gamma   
 *
 * fac_t   the number of vectors to calculate.
 *
 * neq     the number of equations.
 */
{
  /* memory management */
  Complex *Z;
  DistributeWorkspace(Z,sizeof(Complex)*fac_t*KrylovLdSmall(Kry),Work);
  Work = InitKry(Kry,Work,Start,ReBuild);

  Exp4Message1((DEBUG_FILE,"reuse V and H of last rejected step\n"));
  Exp4Message2((DEBUG_FILE,"Krylov start vector || v || = %e, "
	"stored in struct = %e\n",(double)pnrm2(&neq,Start,&inc1),
	(double)KrylovNormv(Kry)));

  /* Calculate z(:,1) = phi(1/3*h*H) * e1;
   *           z(:,2) = phi(2/3*h*H) * e1;
   *           z(:,3) = phi(h*H) * e1; */
  Exp4Message4((DEBUG_FILE,"Calculate z(1:%d,:)\n",KrylovSize(Kry)));
  PhiZ (Kry,Work,KrylovNormv(Kry),fac_t,Z,KrylovLdSmall(Kry),
	  h*gamma,KrylovSize(Kry),KrylovHBandWith(Kry),
	  KrylovH(Kry),KrylovLdSmall(Kry));

  /* Calculate the linear combination of the krylov base vectors with the
     coefitients z. */
  Exp4Message4((DEBUG_FILE,"Doing Linear combiantion now for size = %d\n",
	KrylovSize(Kry)));
  KrylovCombineLinear (X,neq,Start,Kry,Z,KrylovLdSmall(Kry),fac_t,
	0/* don't have old V from build */,1/* force rebuild */);
}


int PREC(PhiKrylovBuild) (WorkSpace Work,Complex *X,finteger LdX,
      krylov *Kry,Complex *v,Real *h,
      Real gamma,finteger fac_t,finteger Guess,
      Real errw,int assr,
      finteger neq)
/* PhiKrylov calculates an approximation to the matrix function phi(h*A)*v.
 * This is done by a krylov subspace aproximation of the krylov space type
 * stored in the krylov structure Kry. The function phi is here 
 * phi(z) = (exp(z)-1)/z.
 *
 * parameters:
 *
 * Work    A pointer to the workspace needed by PhiKrylov
 *
 * x       A pointer to the result matrix including fac_t vectors of the
 *
 * Kry
 *
 * v       the vector to multiply with the matrix function
 *
 * h       the multiplication factor for the matrix A, may be decreased by 
 *         this function if assr is set to Exp4True.
 *
 * gamma   
 *
 * fac_t   the number of vectors to calculate.
 *
 * errw    the actual errw (used for error estimation)
 *
 * assr    flag which indicates if phikry is allowed to reduce the
 *         step size. (assr means Allow Step Size Reduction)
 *
 * neq     the number of equations.
 */
{
  Complex *Z;
#if (defined(SINGLE_PRECISION) || defined(SINGLE_COMPLEX))
  Real res = FLT_MAX;
#else
  Real res = DBL_MAX;
#endif
  int nextind=0;
  int finished=0;
  Real sc_norm = 0.0;
  finteger RealGuessDim = 0;
  /* Flag for use with CGS krylov space,
     if set it is cheaper to use the chached Lanczos vectors in V 
     then increasing the krylov dimension up to RealGuessDim. */
  int AcceptKonvergence = 0;
  int ForceRebuild = 0;

#define z(I,J)      Z[(I) + KrylovLdSmall(Kry)*(J)]

  /* memory management */
  DistributeWorkspace(Z,sizeof(Complex)*fac_t*KrylovLdSmall(Kry),Work);
  /* take the rest of workspace for Phim */

  /**********************************************************/
  /* Generating new Krylov space                            */
  /**********************************************************/

  Exp4Message1((DEBUG_FILE,"building new krylov space\n"));

  /* initialize the krylov array's */
  Work = InitKry(Kry,Work,v,Build);

  Exp4Message2((DEBUG_FILE,"Krylov start vector || v || = %e\n",
	KrylovNormv(Kry)));

  if (KrylovHasHGuess(Kry)) {
    /* for CGS */
    RealGuessDim = KrylovGetHGuess(Work,v,Guess,Kry);
    if (RealGuessDim > 0) {
      Exp4Message3((DEBUG_FILE,"Calculate zGuess(1:%d,:)\n",RealGuessDim));
      /* phiH = phim(full(H(1:k,1:k))); */
      PhiZ (Kry,Work,KrylovNormv(Kry),fac_t,Z,KrylovLdSmall(Kry),
	    *h*gamma,RealGuessDim,KrylovHBandWith(Kry),KrylovH(Kry),
	    KrylovLdSmall(Kry));
    }
  }
  else
    RealGuessDim = 0;

  while(!finished && ((KrylovSize(Kry)) < KrylovMaxSize(Kry)))
  {
    int NextIndex = KrylovIndex(Kry)[nextind];
    NextIndex = min(NextIndex,KrylovMaxSize(Kry));

    /* get a bigger H */
    KrylovIncreaseDimension(Work,&sc_norm,&AcceptKonvergence,X,neq,
	  Z,KrylovLdSmall(Kry),fac_t,v,NextIndex,Kry);
    
    /* calculate the phi(h*H)*e1 */
    if ((RealGuessDim != KrylovSize(Kry)) || (ForceRebuild == 1)) {
      Exp4Message4((DEBUG_FILE,"Calculate z(1:%d,:)\n",KrylovSize(Kry)));
      /* phiH = phim(full(H(1:k,1:k))); */
      PhiZ (Kry,Work,KrylovNormv(Kry),fac_t,Z,KrylovLdSmall(Kry),
	    *h*gamma,KrylovSize(Kry),KrylovHBandWith(Kry),KrylovH(Kry),
	    KrylovLdSmall(Kry));
    }

    /* get the residual */
    res = PREC(GetResidual) (KrylovSize(Kry),*h,gamma,fac_t,sc_norm,errw,
	  Z,KrylovLdSmall(Kry));

    /* if very, very bad residuals are seen reduce the step size */
    if ((KrylovMaxSize(Kry) < KrylovSize(Kry)*2) && assr &&
	  ((TOUGH_BOUND < res) || isnan(res))) {
      Exp4Message1((DEBUG_FILE,"Krylov process detected large res. %7.2e,"
	    " reduce step size by %7.2e\n",res,TOUGH_RED_FAC));
      (*h) *= TOUGH_RED_FAC;
      ForceRebuild = 1;

      Exp4Message4((DEBUG_FILE,"Calculate z(1:%d,:)\n",KrylovSize(Kry)));
      /* phiH = phim(full(H(1:k,1:k))); */
      PhiZ (Kry,Work,KrylovNormv(Kry),fac_t,Z,KrylovLdSmall(Kry),
	    (*h)*gamma,KrylovSize(Kry),KrylovHBandWith(Kry),KrylovH(Kry),
	    KrylovLdSmall(Kry));
      /* get the residual */
      res = PREC(GetResidual) (KrylovSize(Kry),*h,gamma,fac_t,sc_norm,errw,
	    Z,KrylovLdSmall(Kry));
    }

    /* if the error is small enough */
    finished = (res < 1.0);

    /* get the next index from the row of indizes */
    while (KrylovIndex(Kry)[nextind] <= (KrylovSize(Kry))) nextind++;
  }  /* while(!finished && ((KrylovSize(Kry))<KrylovMaxSize(Kry))) */
  KrylovKonvergence(Kry) = KrylovSize(Kry);

  /* complete the guessed krylov dimension if desired */
  if (AcceptKonvergence || ForceRebuild) {
    ForceRebuild = 1;
  } else if (KrylovSize(Kry) < RealGuessDim) {
    /* complete the krylov size up to the dimension RealGuessDim */
    KrylovIncreaseDimension(Work,&sc_norm,&AcceptKonvergence,X,neq,
	  Z,KrylovLdSmall(Kry),fac_t,v,RealGuessDim,Kry);

    /* get the residual */
    res = PREC(GetResidual) (KrylovSize(Kry),*h,gamma,fac_t,sc_norm,errw,
	  Z,KrylovLdSmall(Kry));

    /* if the residual is __NOT__ small enough, take the krylov dimension
       which is konverged */
    if (res >= 1.0) {
      KrylovSize(Kry) = KrylovKonvergence(Kry);
      ForceRebuild = 1;
    }
  }


  if (finished) {
    Exp4Message2((DEBUG_FILE,"Krylov process is finished with %d steps\n",
	  (int)(KrylovSize(Kry))));

    /* Calculate the linear combination of the krylov base vectors with the
       coefitients z. */
    Exp4Message4((DEBUG_FILE,"Doing Linear combiantion now for size = %d\n",
	  KrylovSize(Kry)));
    KrylovCombineLinear (X, neq, v, Kry, Z,
	  KrylovLdSmall(Kry), fac_t, DisableCaching?0:1, ForceRebuild);

    return 0; /* success */
  } else {
    if (assr)/* that means: Allow Step Size Reduction */
    {
      Exp4Message2((DEBUG_FILE,"Krylov process is not finished, must reduce "
	    "step size. h=%e, res = %e\n",(double)(*h),(double)res));

      /* reduce the stepsize as long as the error is not acceptable,
	 or if the lower bound for the stepsize is reached. */
      while (1.0 <= res || isnan(res)) {
	Real red_fac = 1.0;
	Real res_mult = sc_norm*fabs(*h)*gamma*
	  ((Real)fac_t)*max(1.0,errw);

	if (3 <= fac_t) {
	  red_fac=2.0/3.0;
	  if (1.0 <= red_fac * fabs(*h) * res_mult
		* pabs(z(KrylovSize(Kry)-1,1))) {
	    red_fac=1.0/3.0;
	    if (1.0 <= red_fac * fabs(*h) * res_mult
		  * pabs(z(KrylovSize(Kry)-1,0)))
	      red_fac=1.0/8.0;
	  }
	} else {
	  red_fac = 1.0/2.0;
	  if (2 <= fac_t)
	    if (1.0 <= red_fac * fabs(*h) * res_mult
		  * pabs(z(KrylovSize(Kry)-1,1)))
	      red_fac = 1.0/4.0;
	}

        /* reduce the step size */
        (*h) *= red_fac;

	Exp4Message4((DEBUG_FILE,"Calculate z(1:%d,:)\n",KrylovSize(Kry)));
        /* phiH = phim(full(H(1:k,1:k))); */
	PhiZ (Kry,Work,KrylovNormv(Kry),fac_t,Z,KrylovLdSmall(Kry),
	      (*h)*gamma,KrylovSize(Kry),KrylovHBandWith(Kry),
	      KrylovH(Kry),KrylovLdSmall(Kry));
	
	/* get the residual */
	res = PREC(GetResidual) (KrylovSize(Kry),*h,gamma,fac_t,sc_norm,errw,
	      Z,KrylovLdSmall(Kry));

	Exp4Message4((DEBUG_FILE,"Tested residual: reduced step size by %e,"
	      " res = %e\n",(double)red_fac,(double)res));
      } /* while */


      Exp4Message2((DEBUG_FILE,"Krylov process is now converged:"
	    " reduced h = %e, res=%e\n",(double)(*h),(double)res));

      /* Calculate the linear combination of the krylov base vectors with the
	 coefitients z. */
      Exp4Message4((DEBUG_FILE,"Doing Linear combiantion now for size = %d\n",
	    KrylovSize(Kry)));
      KrylovCombineLinear (X, neq, v, Kry, Z, KrylovLdSmall(Kry), fac_t,
	    DisableCaching?0:1, 1);

      return 0; /* success */
    } /* if (assr) */
    else
      /* now we have done all we can do, but anyway,
       * the krylov step failed :-( */
      return 1;
  }
#undef z
}

int PREC(PhiKrylov) (WorkSpace Work,Complex *x,finteger LdX,
      krylov *Kry,Complex *v,Real *h,Real gamma,finteger fac_t,finteger Guess,
      Real errw,int assr,finteger neq)



/* PhiKrylov calculates an approximation to the matrix function phi(h*A)*v.
 * This is done by a krylov subspace aproximation of the krylov space type
 * stored in the krylov structure Kry. The function phi is here 
 * phi(z) = (exp(z)-1)/z.
 *
 * parameters:
 *
 * Work    A pointer to the workspace needed by PhiKrylov
 *
 * x       A pointer to the result matrix including fac_t vectors of the
 *
 * Kry
 *
 * v       the vector to multiply with the matrix function
 *
 * h       the multiplication factor for the matrix A, may be decreased by 
 *         this function if assr is set to Exp4True.
 *
 * gamma   
 *
 * fac_t   the number of vectors to calculate.
 *
 * errw    the actual errw (used for error estimation)
 *
 * assr    flag which indicates if phikry is allowed to reduce the
 *         step size. (assr means Allow Step Size Reduction)
 *
 * neq     the number of equations.
 */
{
  /* reuse existing krylov space */
  if (KrylovSize(Kry) > 0) {
    PREC(PhiKrylovCombineExistingBase) (Work,x,LdX,Kry,v,*h,gamma,fac_t,neq);
    KrylovKonvergence(Kry) = KrylovSize(Kry);
    return 0; /* success */
  } else {
    Real sc_norm;
    
    /* FIXME: Don't use ASSR for that purpose .... */
    /* FIXME: it is thought avoiding all k(:,i) == 0 in case of small h */
    /* FIXME: and small f(y_0) in the first stage while not doing reuse_k! */
    if (!assr) {
      /* if tau * fac_t * scaled norm of v small enough quick return .. */
      sc_norm = PREC (Exp4ScaledNorm) (v);
      if((fabs(*h*gamma)*((Real)fac_t)*sc_norm) < 1e-3) {
	finteger ihelp = fac_t*LdX;
	/* return zero vector. */
	pcopy (&ihelp,&wzero,&inc0,x,&inc1);
	
	Exp4Message1((DEBUG_FILE,"Fast exit from PhiKrylov\n"));
	return 0;/* success */
      }
    
      KrylovNormv(Kry) = pnrm2 (&KrylovNeq(Kry),v,&inc1);
    } else {
      KrylovNormv(Kry) = pnrm2 (&KrylovNeq(Kry),v,&inc1);

      if (KrylovNormv(Kry) == 0.0) {
	finteger ihelp = fac_t*LdX;
	/* return zero vector. */
	pcopy (&ihelp,&wzero,&inc0,x,&inc1);
	
	Exp4Message1((DEBUG_FILE,"Fast exit from PhiKrylov\n"));
	return 0;/* success */
      }
    }

    Exp4Message3((DEBUG_FILE,"In krylov build: || v || = %e\n",
	  KrylovNormv(Kry)));

    if (KrylovHasSecondary(Kry)) {
      int Fail = 1;

      if (KrylovIsPrimary(Kry)) {
	Exp4Message2((DEBUG_FILE,"Try to Build Krylov Space with primary "
	      "struct\n"));

	Fail = PREC(PhiKrylovBuild) (Work,x,LdX,Kry,v,h,gamma,fac_t,Guess,errw,
	    0,neq);
      }

      if (Fail) {
	ClearKrylovSpace(Kry);
	Exp4Message2((DEBUG_FILE,"Try to Build Krylov Space with secondary "
	      "struct\n"));
	KrylovActiveSecondary(Kry);
	Fail = PREC(PhiKrylovBuild) (Work,x,LdX,Kry,v,h,gamma,fac_t,Guess,errw,
	      assr,neq);
      }

      return Fail;
    } else {
      Exp4Message2((DEBUG_FILE,"Try to Build Krylov Space\n"));

      return PREC(PhiKrylovBuild) (Work,x,LdX,Kry,v,h,gamma,fac_t,Guess,errw,
	    assr,neq);
    }
  }
}
